"""Class to wrap PLUMED for use within HyMD.
"""

import numpy as np
import logging
import os
from mpi4py import MPI
from .logger import Logger


class PlumedBias:
    """PLUMED handler class

    Notes
    -----
    This wraps the :code:`Plumed()` class, see
    `https://github.com/plumed/plumed2/tree/master/python`.

    The :code:`PlumedBias()` object is created with the arguments from
    :code:`__init__` and at everystep the methods :code:`prepare()` and
    :code:`calc()` should be called.

    Attributes
    ----------
    plumed_obj : plumed.Plumed
        Plumed object used to pass and request information from PLUMED.
    plumed_forces : (N, D) numpy.ndarray
        Array of forces of :code:`N` particles in :code:`D` dimensions.
        After :code:`calc()`, this array contains only the forces due
        to the PLUMED bias.
    positions : (N, D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
        Needed because we need C-contiguous array for passing to PLUMED.
    plumed_bias : (1,) numpy.ndarray
        Used as a pointer to an :code:`double` to store the bias energy.
    plumed_version : (1,) numpy.ndarray
        Used as a pointer to an :code:`int` to store PLUMED API version.
    comm : mpi4py.Comm
        MPI communicator to use for rank communication.
    ready : bool
        Stores wether the :code:`calc()` method can be called or not.
    """

    plumed_obj = None
    plumed_forces = None
    positions = None
    charges = None
    plumed_bias = np.zeros(1, np.double)
    plumed_version = np.zeros(1, dtype=np.intc)
    comm = None
    ready = False
    verbose = None

    def __init__(self, config, plumeddat, logfile, comm=MPI.COMM_WORLD, verbose=1):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.
        plumeddat : str
            Path file containing PLUMED input.
        logfile : str
            Path to PLUMED's output file
        comm : mpi4py.Comm, optional
            MPI communicator to use for rank communication.
        verbose : int, optional
            Specify the logging event verbosity of this object.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        try:
            import plumed
        except ImportError:
            err_str = (
                "You are trying to use PLUMED " "but HyMD could not import py-plumed."
            )
            Logger.rank0.log(logging.ERROR, err_str)
            raise ImportError(err_str)

        self.comm = comm
        self.verbose = verbose

        try:
            kernel_str = "Using PLUMED_KERNEL={}".format(os.environ["PLUMED_KERNEL"])
        except:
            kernel_str = "The PLUMED_KERNEL environment variable is not set."

        Logger.rank0.log(logging.INFO, kernel_str)

        try:
            self.plumed_obj = plumed.Plumed()
        except:
            err_str = (
                "HyMD was not able to create a PLUMED object. "
                "Maybe there is a problem with your PLUMED_KERNEL?"
            )
            Logger.rank0.log(logging.ERROR, err_str)
            raise RuntimeError(err_str)

        self.plumed_obj.cmd("getApiVersion", self.plumed_version)
        if self.plumed_version[0] <= 3:
            err_str = "HyMD requires a PLUMED API > 3. " "Use a newer PLUMED kernel."
            Logger.rank0.log(logging.ERROR, err_str)
            raise AssertionError(err_str)

        self.plumed_obj.cmd("setMDEngine", "HyMD")
        self.plumed_obj.cmd("setMPIComm", comm)

        Logger.rank0.log(
            logging.INFO, f"Attempting to read PLUMED input from {plumeddat}"
        )

        self.plumed_obj.cmd("setPlumedDat", plumeddat)
        self.plumed_obj.cmd("setLogFile", logfile)
        self.plumed_obj.cmd("setTimestep", config.respa_inner * config.time_step)
        self.plumed_obj.cmd("setNatoms", config.n_particles)
        self.plumed_obj.cmd("setKbT", config.gas_constant * config.target_temperature)
        self.plumed_obj.cmd("setNoVirial")
        self.plumed_obj.cmd("init")

        Logger.rank0.log(
            logging.INFO,
            f"Successfully read PLUMED input. PLUMED output file is {logfile}",
        )

    def finalize(self):
        """Finalize object"""
        self.plumed_obj.finalize()

    def __enter__(self):
        """Allow usage in 'with' context"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Finalize on exit of context"""
        self.plumed_obj.__exit__(exc_type, exc_value, traceback)

    @property
    def api_version(self):
        """Returns the API version got from the PLUMED kernel."""
        return self.plumed_version[0]

    def prepare(self, step, forces, positions, indices, config, charges):
        """Set the pointers to positions and forces, and returns
        wether the potential energy is being requested by PLUMED or not.
        """
        if self.verbose > 1:
            Logger.rank0.log(logging.INFO, f"Setting PLUMED pointers for step {step}")
        self.plumed_forces = forces.ravel(order="C").astype(np.double)
        self.charges = charges.astype(np.double)
        self.positions = positions.ravel(order="C").astype(
            np.double
        )  # get C-contiguous array

        needs_energy = np.zeros(1, np.intc)
        # plumed_virial = np.zeros((3,3), dtype=np.double)
        masses = np.full_like(indices, config.mass, dtype=np.double)
        box = np.diag(config.box_size).astype(np.double)

        self.plumed_obj.cmd("setAtomsNlocal", indices.shape[0])
        self.plumed_obj.cmd("setAtomsGatindex", indices)
        self.plumed_obj.cmd("setStep", step)
        self.plumed_obj.cmd("setForces", self.plumed_forces)
        self.plumed_obj.cmd("setPositions", self.positions)
        self.plumed_obj.cmd("setCharges", self.charges)
        self.plumed_obj.cmd("setMasses", masses)
        self.plumed_obj.cmd("setBox", box)
        # self.plumed_obj.cmd("setVirial", plumed_virial)

        # check if PLUMED needs energy and returns
        self.plumed_obj.cmd("prepareCalc")
        self.plumed_obj.cmd("isEnergyNeeded", needs_energy)

        self.ready = True
        return True if needs_energy[0] != 0 else False

    def calc(self, forces, poteng):
        """Passes the energy (which can be set to any value in case
        :code:`prepare()` returns PLUMED doesn't need it) to get
        the forces and energy from the bias.
        """
        if not self.ready:
            err_str = (
                "Trying to calculate PLUMED biased forces "
                "without first calling prepare method."
            )
            Logger.rank0.log(logging.ERROR, err_str)
            raise RuntimeError(err_str)

        if self.verbose > 1:
            Logger.rank0.log(logging.INFO, "Calculating PLUMED forces")

        # set the energy and calc
        self.plumed_obj.cmd("setEnergy", poteng)
        self.plumed_obj.cmd("performCalc")
        self.plumed_obj.cmd("getBias", self.plumed_bias)

        # check if the returned forces are valid
        if np.isnan(self.plumed_forces).any():
            err_str = (
                "Forces returned by PLUMED are not valid. "
                "It means there's a NaN in the computed forces, "
                "and your input should be checked. "
            )
            Logger.rank0.log(logging.ERROR, err_str)
            raise RuntimeError(err_str)

        # subtract forces to get only the bias' extra force
        self.plumed_forces -= forces.ravel(order="C")

        if self.verbose > 1:
            Logger.rank0.log(logging.INFO, "Done calculating PLUMED forces")

        self.ready = False
        return (
            np.asfortranarray(np.reshape(self.plumed_forces, forces.shape, order="C")),
            self.plumed_bias[0],
        )



from mpi4py import MPI
import pytest
import warnings
import pmesh
import logging
import numpy as np
from hymd.input_parser import Config
from hymd.file_io import distribute_input
from hymd.field import domain_decomposition


@pytest.mark.mpi()
def test_domain_decomposition(molecules_with_solvent, caplog):
    caplog.set_level(logging.INFO)
    indices, positions, molecules, velocities, bonds, names, types = molecules_with_solvent
    box_size = np.array([10, 10, 10], dtype=np.float64)
    config = Config(n_steps=0, time_step=0.03, box_size=box_size,
                    mesh_size=[5, 5, 5], sigma=0.5, kappa=0.05,
                    n_particles=len(indices))
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action='ignore', category=np.VisibleDeprecationWarning,
            message=r'Creating an ndarray from ragged nested sequences'
        )
        pm = pmesh.ParticleMesh(config.mesh_size, BoxSize=config.box_size,
                                dtype='f4', comm=comm)

    # Test stub acting like a hdf5 file for distribute_input
    in_file = {'molecules': molecules, 'indices': indices}
    rank_range, molecules_flag = distribute_input(
        in_file, rank, size,  config.n_particles, 6, comm=comm
    )
    positions_ = positions[rank_range]
    molecules_ = molecules[rank_range]
    indices_ = indices[rank_range]
    velocities_ = velocities[rank_range]
    bonds_ = bonds[rank_range]
    types_ = types[rank_range]
    names_ = names[rank_range]

    dd = domain_decomposition(
        positions_,
        pm,
        indices_,
        velocities_,
        names_,
        types_,
        molecules=molecules_,
        bonds=bonds_,
        verbose=2,
        comm=comm
    )
    positions_, indices_, velocities_, names_, types_, bonds_, molecules_ = dd
    if rank == 0:
        assert 'DOMAIN_DECOMP' in caplog.text

    unique_molecules = np.unique(molecules_)
    receive_buffer = comm.gather(unique_molecules, root=0)
    unique_molecules = comm.bcast(receive_buffer, root=0)

    # Ensure no molecules are split across ranks
    for i, ranki in enumerate(unique_molecules):
        for mi in ranki:
            for j, rankj in enumerate(unique_molecules):
                if i != j:
                    assert mi not in rankj



"""Interaction energy functionals.

Implements field--particle interaction energy functionals depending on the
field configuration. Extending HyMD with a new Hamiltonian is done by
subclassing the :code:`Hamiltonian` superclass with logic intialized in the
constructor.

The grid-independent filtering function is defined in the :code:`Hamiltonian`
superclass constructor. Hijack :code:`self.H` in a subclass to change the
filter.
"""
import numpy as np
import sympy


class Hamiltonian:
    """Interaction energy functional superclass"""

    def __init__(self, config):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        self.config = config
        self._setup()

    def _setup(self):
        """Superclass setup

        Sets up the grid-independent filtering function :code:`H`, and the
        SymPy logic for symbolically differentiating interaction energy
        functionals in Hamiltonian subclasses.
        """
        if not hasattr(self.config, "simulation_volume"):
            self.config.simulation_volume = np.prod(np.asarray(self.config.box_size))
        if not self.config.barostat:
            self.config.rho0 = self.config.n_particles / self.config.simulation_volume
            self.config.a = self.config.rho0
        if not self.config.rho0:
            self.config.rho0 = self.config.n_particles / self.config.simulation_volume
        self.phi = sympy.var("phi:%d" % (self.config.n_types))
        k = sympy.var("k:%d" % (3))

        # electrostatics variables
        self.psi = sympy.var("psi")
        self.phi_q = sympy.var("phi_q")
        if not self.config.self_energy:
            self.config.self_energy = 0.0

        def fourier_space_window_function(k):
            return sympy.functions.elementary.exponential.exp(
                -0.5
                * self.config.sigma**2
                * (k0**2 + k1**2 + k2**2)  # noqa: F821, E501
            )

        self.window_function_lambda = sympy.lambdify(
            [k], fourier_space_window_function(k)
        )

        def H(k, v):
            return v * self.window_function_lambda(k)

        self.H = H


class SquaredPhi(Hamiltonian):
    """Simple squared density interaction energy functional

    The interaction energy density takes the form

    .. math::

        w[\\tilde\\phi] = \\frac{1}{2\\kappa\\rho_0}
            \\left(
                \\sum_k \\tilde\\phi_k
            \\right)^2,

    where :math:`\\kappa` is the compressibility and :math:`rho_0` is the
    average density of the fully homogenous system. Expressing the species
    densities in terms of fluctuations from the average,

    .. math::

        \\mathrm{d}\\tilde\\phi_k = \\rho_0 - \\tilde\\phi_k,

    it is evident that this interaction energy functional is a slight change of
    the :code:`DefaultNoChi` Hamiltonian, as the expanded interaction energy
    density becomes

    .. math::

        w[\\tilde\\phi] = \\frac{1}{2\\kappa\\rho_0}
            \\left(
                6\\rho_0\\left[
                    \\sum_k \\mathrm{d}\\tilde\\phi_k
                \\right]
                +
                9\\rho_0^2
                +
                \\sum_k \\mathrm{d}\\tilde\\phi_k^2
                +
                \\prod_{k\\not=l}
                    \\mathrm{d}\\tilde\\phi_k \\mathrm{d}\\tilde\\phi_l
            \\right),

    identical to :code:`DefaultNoChi` apart from a constant energy shift
    :math:`9\\rho_0^2`(which does not impact dynamics) and the
    :math:`\\rho_0`--:math:`\\mathrm{d}\\tilde\\phi_k` cross-terms. These
    cross-terms constitute contributions to the energy linear in
    :math:`\\mathrm{d}\\tilde\\phi_k` absent in :code:`DefaultNoChi`, which has
    only quadratic terms present.

    See also
    --------
    hymd.hamiltonian.DefaultNoChi
    """

    def __init__(self, config):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        super(SquaredPhi, self).__init__(config)
        super(SquaredPhi, self)._setup()
        self.setup()

    def setup(self):
        """Setup the interaction energy potential and the external potential"""

        def w(phi, kappa=self.config.kappa, rho0=self.config.rho0):
            return 0.5 / (kappa * rho0) * (sum(phi)) ** 2

        def w_elec(
            phi_q,
            psi,
            volume=self.config.simulation_volume,
            self_energy=self.config.self_energy,
        ):
            self_energy /= volume
            return 0.5 * phi_q * psi - self_energy

        def V_bar_0(
            phi,
            k,
            kappa=self.config.kappa,
            rho0=self.config.rho0,
        ):
            V_incompressibility = 1 / (kappa * rho0) * sum(phi)
            V_interaction = 0
            return V_interaction + V_incompressibility

        def V_bar_elec(
            psi,
            t,
            type_charges=self.config.type_charges,
        ):
            return type_charges[t] * psi

        self.V_bar_0 = [
            sympy.lambdify(
                [(self.phi)], V_bar_0(self.phi, t)
            )
            for t in range(self.config.n_types)
        ]

        self.V_bar = [
            sympy.lambdify(
                [(self.phi, self.psi)], V_bar_0(self.phi, t) + V_bar_elec(self.psi, t)
            )
            for t in range(self.config.n_types)
        ]

        self.v_ext = [
            sympy.lambdify([self.phi], sympy.diff(w(self.phi), self.phi[i]))
            for i in range(self.config.n_types)
        ]

        self.w_0 = sympy.lambdify([self.phi], w(self.phi))
        self.w_elec = sympy.lambdify(
            [(self.phi_q, self.psi)], w_elec(self.phi_q, self.psi)
        )

        if self.config.coulombtype == "PIC_Spectral":
            self.w = sympy.lambdify(
                [(self.phi, self.phi_q, self.psi)],
                w(self.phi) + w_elec(self.phi_q, self.psi),
            )
        else:
            self.w = self.w_0


class DefaultNoChi(Hamiltonian):
    """Incompressibility-only interaction energy functional

    The interaction energy density takes the form

    .. math::

        w[\\tilde\\phi] = \\frac{1}{2\\kappa} \\left(
            \\sum_k \\tilde\\phi_k - a
        \\right)^2,

    where :math:`\\kappa` is the compressibility and :math:`a=\\rho_0` for
    NVT runs where :math:`\\rho_0` is the average density of the fully
    homogenous system. In case of NPT runs, :math:`a` is a calibrated
    parameter to obtain the correct average density at the target temperature
    and pressure. The :code:`SquaredPhi` Hamiltonian implements a similar
    functional with an additional linear term component depending on

    .. math::

        \\mathmr{d}\\tilde\\phi_k = \\tilde\\phi_k - \\rho_0

    and not :math:`\\mathrm{d}\\tilde\\phi_k^2`. No explicit inter-species
    interaction is present apart from the indirect interaction through the
    incompressibility.

    See also
    --------
    hymd.hamiltonian.DefaultNoChi
    hymd.input_parser.Config :
        Configuration dataclass handler.
    """

    def __init__(self, config):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        super(DefaultNoChi, self).__init__(config)
        super(DefaultNoChi, self)._setup()
        self.setup()

    def setup(self):
        """Setup the interaction energy potential and the external potential"""

        def w(phi, kappa=self.config.kappa, rho0=self.config.rho0, a=self.config.a):
            return 0.5 / (kappa * rho0) * (sum(phi) - a) ** 2

        def w_elec(
            phi_q,
            psi,
            volume=self.config.simulation_volume,
            self_energy=self.config.self_energy,
        ):
            self_energy /= volume
            return 0.5 * phi_q * psi - self_energy

        def V_bar_0(
            phi,
            k,
            kappa=self.config.kappa,
            rho0=self.config.rho0,
            a=self.config.a,
        ):
            V_incompressibility = 1 / (kappa * rho0) * (sum(phi) - a)
            V_interaction = 0
            return V_interaction + V_incompressibility

        def V_bar_elec(
            psi,
            t,
            type_charges=self.config.type_charges,
        ):
            return type_charges[t] * psi

        self.V_bar_0 = [
            sympy.lambdify(
                [(self.phi)], V_bar_0(self.phi, t)
            )
            for t in range(self.config.n_types)
        ]

        self.V_bar = [
            sympy.lambdify(
                [(self.phi, self.psi)], V_bar_0(self.phi, t) + V_bar_elec(self.psi, t)
            )
            for t in range(self.config.n_types)
        ]

        self.v_ext = [
            sympy.lambdify([self.phi], sympy.diff(w(self.phi), self.phi[i]))
            for i in range(self.config.n_types)
        ]

        self.w_0 = sympy.lambdify([self.phi], w(self.phi))
        self.w_elec = sympy.lambdify(
            [(self.phi_q, self.psi)], w_elec(self.phi_q, self.psi)
        )

        if self.config.coulombtype == "PIC_Spectral":
            self.w = sympy.lambdify(
                [(self.phi, self.phi_q, self.psi)],
                w(self.phi) + w_elec(self.phi_q, self.psi),
            )
        else:
            self.w = self.w_0


class DefaultWithChi(Hamiltonian):
    """Incompressibility and :math:`\\chi`-interactions energy functional

    The interaction energy density takes the form

    .. math::

        w[\\tilde\\phi] =
            \\frac{1}{2\\rho_0}
                \\sum_{k,l}\\chi_{kl} \\tilde\\phi_k \\tilde\\phi_l
            +
            \\frac{1}{2\\kappa} \\left(
                \\sum_k \\tilde\\phi_k - a
            \\right)^2,

    where :math:`\\kappa` is the compressibility and :math:`a=\\rho_0` for
    NVT runs where :math:`\\rho_0` is the average density of the fully
    homogenous system. In case of NPT runs, :math:`a` is a calibrated
    parameter to obtain the correct average density at the target temperature
    and pressure. :math:`\\chi_{ij}` is the Flory-Huggins-like
    inter-species mixing energy.
    """

    def __init__(self, config, unique_names, type_to_name_map):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.
        unique_names : numpy.ndarray
            Sorted array of all distinct names of different species present in
            the simulation. Result of :code:`numpy.unique(all_names)`, where
            :code:`all_names` is the gathered union of all individual MPI
            ranks' :code:`names` arrays.
        type_to_name_map : dict[int, str]
            Dictionary of the mapping from type indices (integers) to type
            names.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        super(DefaultWithChi, self).__init__(config)
        super(DefaultWithChi, self)._setup()
        self.setup(unique_names, type_to_name_map)

    def setup(self, unique_names, type_to_name_map):
        """Setup the interaction energy potential and the external potential

        Parameters
        ----------
        unique_names : numpy.ndarray
            Sorted array of all distinct names of different species present in
            the simulation. Result of :code:`numpy.unique(all_names)`, where
            :code:`all_names` is the gathered union of all individual MPI
            ranks' :code:`names` arrays.
        type_to_name_map : dict[int, str]
            Dictionary of the mapping from type indices (integers) to type
            names.
        """
        self.type_to_name_map = type_to_name_map
        self.chi_type_dictionary = {
            tuple(sorted([c.atom_1, c.atom_2])): c.interaction_energy
            for c in self.config.chi
        }
        self.phi_laplacian = [
            sympy.var("phi_laplacian%d(0:%d)" % (t, 3))
            for t in range(self.config.n_types)
        ]

        def w(
            phi,
            kappa=self.config.kappa,
            rho0=self.config.rho0,
            a=self.config.a,
            chi=self.config.chi,
            type_to_name_map=self.type_to_name_map,
            chi_type_dictionary=self.chi_type_dictionary,
        ):
            interaction = 0.0
            for i in range(self.config.n_types):
                for j in range(i + 1, self.config.n_types):
                    ni = type_to_name_map[i]
                    nj = type_to_name_map[j]
                    names = sorted([ni, nj])
                    c = chi_type_dictionary[tuple(names)]

                    interaction += c * phi[i] * phi[j] / rho0
            incompressibility = 0.5 / (kappa * rho0) * (sum(phi) - a) ** 2
            return incompressibility + interaction

        def w_elec(
            phi_q,
            psi,
            volume=self.config.simulation_volume,
            self_energy=self.config.self_energy,
        ):
            self_energy /= volume
            return 0.5 * phi_q * psi - self_energy

        def V_bar_0(
            phi,
            t,
            kappa=self.config.kappa,
            rho0=self.config.rho0,
            a=self.config.a,
            chi=self.config.chi,
            type_to_name_map=self.type_to_name_map,
            chi_type_dictionary=self.chi_type_dictionary,
        ):
            V_incompressibility = 1 / (kappa * rho0) * (sum(phi) - a)

            V_interaction = 0.0
            nk = type_to_name_map[t]
            for i in range(self.config.n_types):
                ni = type_to_name_map[i]
                names = sorted([nk, ni])
                if ni != nk:
                    c = chi_type_dictionary[tuple(names)]
                else:
                    c = 0.0
                # uncomment to count diagonal chi terms:
                # c = chi_type_dictionary[tuple(names)]
                V_interaction += c * phi[i] / rho0
            return V_interaction + V_incompressibility

        def V_bar_elec(
            psi,
            t,
            type_charges=self.config.type_charges,
        ):
            return type_charges[t] * psi

        self.V_bar_0 = [
            sympy.lambdify(
                [(self.phi)], V_bar_0(self.phi, t)
            )
            for t in range(self.config.n_types)
        ]

        self.V_bar = [
            sympy.lambdify(
                [(self.phi, self.psi)], V_bar_0(self.phi, t) + V_bar_elec(self.psi, t)
            )
            for t in range(self.config.n_types)
        ]

        self.v_ext = [
            sympy.lambdify([self.phi], sympy.diff(w(self.phi), self.phi[t]))
            for t in range(self.config.n_types)
        ]

        self.w_0 = sympy.lambdify([self.phi], w(self.phi))
        self.w_elec = sympy.lambdify(
            [(self.phi_q, self.psi)], w_elec(self.phi_q, self.psi)
        )

        if self.config.coulombtype == "PIC_Spectral":
            self.w = sympy.lambdify(
                [(self.phi, self.phi_q, self.psi)],
                w(self.phi) + w_elec(self.phi_q, self.psi),
            )
        else:
            self.w = self.w_0


def get_hamiltonian(config):
    """Return appropriate Hamiltonian object based on the
    config.hamiltonian string.

    Parameters
    ----------
    config : Config
        Configuration object.

    Returns
    ----------
    hamiltonian : Hamiltonian
        Hamiltonian object.
    """
    if config.hamiltonian.lower() == "defaultnochi":
        hamiltonian = DefaultNoChi(config)
    elif config.hamiltonian.lower() == "defaultwithchi":
        hamiltonian = DefaultWithChi(
            config, config.unique_names, config.type_to_name_map
        )
    elif config.hamiltonian.lower() == "squaredphi":
        hamiltonian = SquaredPhi(config)

    return hamiltonian



import numba
import numpy as np
from mpi4py import MPI
from compute_gaussian_core import cgc as gaussian_core_kernel__fortran


def setup_chi_array(positions, types, config):
    N = len(positions)
    chi = np.empty(shape=(N, N), dtype=np.float32, order="F")

    chi_type = np.empty(shape=(5, 5), dtype=np.float32)
    for c in config.chi:
        t1 = config.name_to_type_map[c.atom_1]
        t2 = config.name_to_type_map[c.atom_2]
        e = c.interaction_energy
        chi_type[t1, t2] = e
        chi_type[t2, t1] = e

    @numba.jit(nopython=True)
    def chi_array_kernel(chi_type, chi):
        for i in range(N):
            for j in range(i + 1, N):
                chi[i, j] = chi_type[types[i], types[j]]
                chi[j, i] = chi[i, j]
        return chi

    chi = chi_array_kernel(chi_type, chi)
    return chi


@numba.jit(nopython=True, fastmath=True)
def gaussian_core_kernel__numba(r, chi, f, box_size, sigma, kappa):
    f.fill(0.0)
    denominator = 1 / (4 * sigma ** 2)
    V = np.prod(box_size)
    phi0 = len(r) / V
    factor = 1 / (16 * np.pi ** (3 / 2) * kappa * sigma ** 5 * phi0)

    for i, ri in enumerate(r):
        for j in range(i + 1, len(r)):
            rj = r[j, :]
            rij = rj - ri

            # Apply periodic boundary conditions to the distance rij
            rij[0] -= box_size[0] * np.around(rij[0] / box_size[0])
            rij[1] -= box_size[1] * np.around(rij[1] / box_size[1])
            rij[2] -= box_size[2] * np.around(rij[2] / box_size[2])

            dr2 = np.dot(rij, rij)
            exp = np.exp(-dr2 * denominator)

            f_bond_vector = rij * exp * factor * (1 + kappa * chi[i, j])
            f[i, :] += f_bond_vector
            f[j, :] -= f_bond_vector


def gaussian_core_forces(positions, force, chi, config):
    energy = gaussian_core_kernel__fortran(
        positions,
        chi,
        force,
        np.asarray(config.box_size),
        config.sigma,
        config.kappa,
        len(positions),
    )
    return energy


def __setup_test_dppc_system(N_lipid, box=10.0, seed=None):
    from force import Chi
    from input_parser import Config, _setup_type_to_name_map, _find_unique_names

    if seed is not None:
        np.random.seed(seed)
    ratio = 2.2095959596
    N_solvent = int(round(N_lipid * 12 * ratio))
    N = int(round(12 * N_lipid + N_solvent))
    positions = np.random.uniform(low=0.0, high=box, size=(N, 3))
    positions = np.asfortranarray(positions).astype(np.float32)
    forces = np.zeros_like(positions, dtype=np.float32)
    dppc_types = np.array([0, 1, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3], dtype=np.int32)
    dppc_names = np.array(
        [
            np.string_(s)
            for s in ("N", "P", "G", "G", "C", "C", "C", "C", "C", "C", "C", "C")
        ],
        dtype="S1",
    )
    names = np.concatenate(
        (
            np.repeat(dppc_names, N_lipid),
            np.full(N_solvent, np.string_("W"), dtype="S1"),
        )
    )
    types = np.concatenate(
        (np.repeat(dppc_types, N_lipid), np.full(N_solvent, 4, dtype=np.int32))
    )
    chi_list = [
        Chi(atom_1="C", atom_2="W", interaction_energy=42.24),
        Chi(atom_1="G", atom_2="C", interaction_energy=10.47),
        Chi(atom_1="N", atom_2="W", interaction_energy=-3.77),
        Chi(atom_1="G", atom_2="W", interaction_energy=4.53),
        Chi(atom_1="N", atom_2="P", interaction_energy=-9.34),
        Chi(atom_1="P", atom_2="G", interaction_energy=8.04),
        Chi(atom_1="N", atom_2="G", interaction_energy=1.97),
        Chi(atom_1="P", atom_2="C", interaction_energy=14.72),
        Chi(atom_1="P", atom_2="W", interaction_energy=-1.51),
        Chi(atom_1="N", atom_2="C", interaction_energy=13.56),
    ]
    for n in ("N", "P", "G", "C", "W"):
        chi_list.append(Chi(atom_1=n, atom_2=n, interaction_energy=0.0))

    config = Config(
        n_steps=None,
        time_step=None,
        box_size=[box, box, box],
        mesh_size=[3, 3, 3],
        sigma=0.5,
        kappa=0.05,
        chi=chi_list,
        n_particles=N,
    )
    config = _setup_type_to_name_map(config, names, types)
    config = _find_unique_names(config, names)
    return positions, forces, names, types, config


def __force_energy_grid_size_accuracy():
    import warnings
    import pmesh.pm as pmesh
    import matplotlib.pyplot as plt
    from matplotlib.ticker import NullFormatter
    from file_io import distribute_input
    from field import update_field, compute_field_force
    from hamiltonian import DefaultWithChi

    r, f, names, types, config = __setup_test_dppc_system(8 * 66, box=10.0, seed=27652)
    N = config.n_particles
    chi = setup_chi_array(r, types, config)

    if MPI.COMM_WORLD.Get_rank() == 0:
        yticks = [0.001, 0.01, 0.1, 1.0, 10.0, 100.0]
        yticklabels = ["0.001", "0.01", "0.1", "1", "10", "100"]
        xticks = [0.05, 0.01, 0.1, 0.5, 1, 2, 3]
        xticklabels = ["0.05", "0.01", "0.1", "0.5", "1", "2", "3"]
        fig1, ax1 = plt.subplots(constrained_layout=True)
        fig2, ax2 = plt.subplots(constrained_layout=True)
        for ax, title in zip((ax1, ax2), ("force", "energy")):
            ax.set_xlabel("grid length, h [nm]")
            ax.set_ylabel("100% · |GC - PMESH| / |GC|")
            ax.set(title=title, xscale="log", yscale="log")
            ax.yaxis.set_minor_formatter(NullFormatter())
            ax.xaxis.set_minor_formatter(NullFormatter())
            ax.set_yticks(yticks)
            ax.set_yticklabels(yticklabels)
            ax.set_xticks(xticks)
            ax.set_xticklabels(xticklabels)
            props = dict(boxstyle="round", facecolor="gray", alpha=0.1)
            ax.text(
                0.05,
                0.95,
                f"density {N/(config.box_size[0]**3):.2f}",
                transform=ax.transAxes,
                verticalalignment="top",
                bbox=props,
            )
            ax.grid()

    for i, sigma in enumerate([0.25, 0.5, 0.75, 1.0]):
        sigma_str = f"{sigma:.2f}"
        config.sigma = sigma
        gaussian_core_energy = gaussian_core_forces(r, f, chi, config)

        in_file = {"positions": r, "indices": np.arange(config.n_particles)}
        grab, _ = distribute_input(
            in_file,
            MPI.COMM_WORLD.Get_rank(),
            MPI.COMM_WORLD.Get_size(),
            config.n_particles,
            max_molecule_size=15,
            comm=MPI.COMM_WORLD,
        )
        pos_ = r[grab, :]
        typ_ = types[grab]
        mesh = np.geomspace(3, 200, num=50, dtype=np.int32)
        mesh = np.unique(mesh)

        ind = np.random.randint(len(pos_))
        if MPI.COMM_WORLD.Get_rank() == 0:
            print(
                f' mesh{106*" "}relative energy\n size     {"force":>32s} '
                f'{"energy/N":>20s} {"relative force difference [%]":>43s} '
                f'    {"difference [%]":>12s}\n{126*"="}'
            )
            print(
                f"     {f[ind, 0]:12.8f} {f[ind, 1]:12.8f} {f[ind, 2]:12.8f}"
                f" {gaussian_core_energy / N:20.10f}"
            )

        energy_relative_errors = np.zeros_like(mesh, dtype=np.float64)
        energy_absolute_errors = np.zeros_like(mesh, dtype=np.float64)
        force_relative_errors = np.zeros(
            shape=(len(mesh), 3), dtype=np.float64
        )  # noqa: E501
        force_absolute_errors = np.zeros(
            shape=(len(mesh), 3), dtype=np.float64
        )  # noqa: E501

        for i, m in enumerate(mesh):
            config.mesh_size = [m, m, m]
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    category=np.VisibleDeprecationWarning,
                    message=r"Creating an ndarray from ragged nested sequences",
                )
                pm = pmesh.ParticleMesh(
                    config.mesh_size,
                    BoxSize=config.box_size,
                    dtype="f4",
                    comm=MPI.COMM_WORLD,
                )
            phi = [pm.create("real", value=0.0) for _ in range(config.n_types)]
            phi_fourier = [
                pm.create("complex", value=0.0) for _ in range(config.n_types)
            ]
            force_on_grid = [
                [pm.create("real", value=0.0) for d in range(3)]
                for _ in range(config.n_types)
            ]
            v_ext_fourier = [pm.create("complex", value=0.0) for _ in range(4)]
            v_ext = [pm.create("real", value=0.0) for _ in range(config.n_types)]
            layouts = [pm.decompose(pos_[typ_ == t]) for t in range(config.n_types)]
            hamiltonian = DefaultWithChi(
                config, config.unique_names, config.type_to_name_map
            )
            update_field(
                phi,
                layouts,
                force_on_grid,
                hamiltonian,
                pm,
                pos_,
                typ_,
                config,
                v_ext,
                phi_fourier,
                v_ext_fourier,
                compute_potential=False,
            )
            field_forces = np.zeros_like(pos_)
            compute_field_force(
                layouts, pos_, force_on_grid, field_forces, typ_, config.n_types
            )
            V = np.prod(config.box_size)
            n_mesh__cells = np.prod(np.full(3, config.mesh_size))
            volume_per_cell = V / n_mesh__cells
            w = hamiltonian.w(phi) * volume_per_cell
            field_energy = w.csum()

            for d in range(3):
                f_gc = f[grab, d]
                f_pm = field_forces[:, d]
                safe_ind = np.invert(np.isclose(f_gc, 0.0, atol=0.01))
                diff_f = np.abs(f_gc[safe_ind] - f_pm[safe_ind])
                force_absolute_errors[i, d] = np.mean(diff_f)
                force_relative_errors[i, d] = np.mean(diff_f / np.abs(f_gc[safe_ind]))
                force_absolute_errors[i, d] = MPI.COMM_WORLD.allreduce(
                    force_absolute_errors[i, d], MPI.SUM
                )
                force_absolute_errors[i, d] /= MPI.COMM_WORLD.Get_size()
                force_relative_errors[i, d] = MPI.COMM_WORLD.allreduce(
                    force_relative_errors[i, d], MPI.SUM
                )
                force_relative_errors[i, d] /= MPI.COMM_WORLD.Get_size()

            diff_e = np.abs(gaussian_core_energy - field_energy)
            energy_relative_errors[i] = diff_e / gaussian_core_energy
            energy_absolute_errors[i] = diff_e

            if MPI.COMM_WORLD.Get_rank() == 0:
                ff = field_forces
                df = np.abs(ff[ind, :] - f[ind, :]) / np.abs(f[ind, :]) * 100.0
                de = (
                    np.abs(gaussian_core_energy - field_energy)
                    / np.abs(gaussian_core_energy)
                ) * 100.0
                print(
                    f"{m:4} {ff[ind, 0]:12.8f} {ff[ind, 1]:12.8f} "
                    f"{ff[ind, 2]:12.8f} {field_energy / N:20.10f}     "
                    f"{df[0]:12.8f} {df[1]:12.8f} {df[2]:12.8f} {de:17.8f}"
                )

        if MPI.COMM_WORLD.Get_rank() == 0:
            ax1.plot(
                config.box_size[0] / mesh,
                100.0 * np.mean(force_relative_errors, axis=1),
                "-",
                label=f"force σ={sigma_str}",
            )
            ax2.plot(
                config.box_size[0] / mesh,
                100.0 * energy_relative_errors,
                "-",
                label=f"energy σ={sigma_str}",
            )

    if MPI.COMM_WORLD.Get_rank() == 0:
        ax1.legend(loc="lower right")
        ax2.legend(loc="lower right")
        plt.show()


def __force_energy_box_size_sigma_accuracy():
    from math import ceil
    import warnings
    import pmesh.pm as pmesh
    import matplotlib.pyplot as plt
    from matplotlib.ticker import NullFormatter
    from file_io import distribute_input
    from field import update_field, compute_field_force
    from hamiltonian import DefaultWithChi

    sigma = 0.3

    if MPI.COMM_WORLD.Get_rank() == 0:
        yticks = [0.001, 0.01, 0.1, 1.0, 10.0, 100.0]
        yticklabels = ["0.001", "0.01", "0.1", "1", "10", "100"]
        xticks = [0.05, 0.01, 0.1, 0.5, 1, 2, 3]
        xticklabels = ["0.05", "0.01", "0.1", "0.5", "1", "2", "3"]
        fig1, ax1 = plt.subplots(constrained_layout=True)
        fig2, ax2 = plt.subplots(constrained_layout=True)
        for ax, title in zip((ax1, ax2), ("force", "energy")):
            ax.set_xlabel("grid length, h [nm]")
            ax.set_ylabel("100% · |GC - PMESH| / |GC|")
            ax.set(title=title, xscale="log", yscale="log")
            ax.yaxis.set_minor_formatter(NullFormatter())
            ax.xaxis.set_minor_formatter(NullFormatter())
            ax.set_yticks(yticks)
            ax.set_yticklabels(yticklabels)
            ax.set_xticks(xticks)
            ax.set_xticklabels(xticklabels)
            props = dict(boxstyle="round", facecolor="gray", alpha=0.1)
            ax.text(
                0.05,
                0.95,
                f"sigma {sigma:.3f}",
                transform=ax.transAxes,
                verticalalignment="top",
                bbox=props,
            )
            ax.grid()

    for i, box_ in enumerate(np.arange(4, 15, 2).astype(np.float64)):
        box = box_ * sigma
        lipid_density = 2.6779374472
        N_lipids = int(ceil(lipid_density * box ** 3 / 12))
        r, f, names, types, config = __setup_test_dppc_system(
            N_lipids, box=box, seed=27652
        )
        N = config.n_particles
        config.sigma = sigma
        chi = setup_chi_array(r, types, config)
        gaussian_core_energy = gaussian_core_forces(r, f, chi, config)

        in_file = {"positions": r, "indices": np.arange(config.n_particles)}
        grab, _ = distribute_input(
            in_file,
            MPI.COMM_WORLD.Get_rank(),
            MPI.COMM_WORLD.Get_size(),
            config.n_particles,
            max_molecule_size=15,
            comm=MPI.COMM_WORLD,
        )
        pos_ = r[grab, :]
        typ_ = types[grab]
        mesh = np.geomspace(3, int(round(box / 0.01)), num=50, dtype=np.int32)
        # mesh = np.arange(3, 200, dtype=np.int32)
        mesh = np.unique(mesh)

        ind = np.random.randint(len(pos_))
        if MPI.COMM_WORLD.Get_rank() == 0:
            print(
                f' mesh{106*" "}relative energy\n size     {"force":>32s} '
                f'{"energy/N":>20s} {"relative force difference [%]":>43s} '
                f'    {"difference [%]":>12s}\n{126*"="}'
            )
            print(
                f"     {f[ind, 0]:12.8f} {f[ind, 1]:12.8f} {f[ind, 2]:12.8f}"
                f" {gaussian_core_energy / N:20.10f}"
            )

        energy_relative_errors = np.zeros_like(mesh, dtype=np.float64)
        energy_absolute_errors = np.zeros_like(mesh, dtype=np.float64)
        force_relative_errors = np.zeros(
            shape=(len(mesh), 3), dtype=np.float64
        )  # noqa: E501
        force_absolute_errors = np.zeros(
            shape=(len(mesh), 3), dtype=np.float64
        )  # noqa: E501

        for i, m in enumerate(mesh):
            config.mesh_size = [m, m, m]
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    category=np.VisibleDeprecationWarning,
                    message=r"Creating an ndarray from ragged nested sequences",
                )
                pm = pmesh.ParticleMesh(
                    config.mesh_size,
                    BoxSize=config.box_size,
                    dtype="f4",
                    comm=MPI.COMM_WORLD,
                )
            phi = [pm.create("real", value=0.0) for _ in range(config.n_types)]
            phi_fourier = [
                pm.create("complex", value=0.0) for _ in range(config.n_types)
            ]
            force_on_grid = [
                [pm.create("real", value=0.0) for d in range(3)]
                for _ in range(config.n_types)
            ]
            v_ext_fourier = [pm.create("complex", value=0.0) for _ in range(4)]
            v_ext = [pm.create("real", value=0.0) for _ in range(config.n_types)]
            layouts = [pm.decompose(pos_[typ_ == t]) for t in range(config.n_types)]
            hamiltonian = DefaultWithChi(
                config, config.unique_names, config.type_to_name_map
            )
            update_field(
                phi,
                layouts,
                force_on_grid,
                hamiltonian,
                pm,
                pos_,
                typ_,
                config,
                v_ext,
                phi_fourier,
                v_ext_fourier,
                compute_potential=False,
            )
            field_forces = np.zeros_like(pos_)
            compute_field_force(
                layouts, pos_, force_on_grid, field_forces, typ_, config.n_types
            )
            V = np.prod(config.box_size)
            n_mesh__cells = np.prod(np.full(3, config.mesh_size))
            volume_per_cell = V / n_mesh__cells
            w = hamiltonian.w(phi) * volume_per_cell
            field_energy = w.csum()

            for d in range(3):
                f_gc = f[grab, d]
                f_pm = field_forces[:, d]
                safe_ind = np.invert(np.isclose(f_gc, 0.0, atol=0.01))
                diff_f = np.abs(f_gc[safe_ind] - f_pm[safe_ind])
                force_absolute_errors[i, d] = np.mean(diff_f)
                force_relative_errors[i, d] = np.mean(diff_f / np.abs(f_gc[safe_ind]))
                force_absolute_errors[i, d] = MPI.COMM_WORLD.allreduce(
                    force_absolute_errors[i, d], MPI.SUM
                )
                force_absolute_errors[i, d] /= MPI.COMM_WORLD.Get_size()
                force_relative_errors[i, d] = MPI.COMM_WORLD.allreduce(
                    force_relative_errors[i, d], MPI.SUM
                )
                force_relative_errors[i, d] /= MPI.COMM_WORLD.Get_size()

            diff_e = np.abs(gaussian_core_energy - field_energy)
            energy_relative_errors[i] = diff_e / gaussian_core_energy
            energy_absolute_errors[i] = diff_e

            if MPI.COMM_WORLD.Get_rank() == 0:
                ff = field_forces
                df = np.abs(ff[ind, :] - f[ind, :]) / np.abs(f[ind, :]) * 100.0
                de = (
                    np.abs(gaussian_core_energy - field_energy)
                    / np.abs(gaussian_core_energy)
                ) * 100.0
                print(
                    f"{m:4} {ff[ind, 0]:12.8f} {ff[ind, 1]:12.8f} "
                    f"{ff[ind, 2]:12.8f} {field_energy / N:20.10f}     "
                    f"{df[0]:12.8f} {df[1]:12.8f} {df[2]:12.8f} {de:17.8f}"
                )

        if MPI.COMM_WORLD.Get_rank() == 0:
            ax1.plot(
                config.box_size[0] / mesh,
                100.0 * np.mean(force_relative_errors, axis=1),
                "-",
                label=f"box={box_:.0f}σ",
            )
            ax2.plot(
                config.box_size[0] / mesh,
                100.0 * energy_relative_errors,
                "-",
                label=f"box={box_:.0f}σ",
            )

    if MPI.COMM_WORLD.Get_rank() == 0:
        ax1.legend(loc="lower right")
        ax2.legend(loc="lower right")
        plt.show()


if __name__ == "__main__":
    # __force_energy_grid_size_accuracy()
    __force_energy_box_size_sigma_accuracy()



import numpy as np
import h5py
from hymd.file_io import distribute_input


def _distribute_nranks(size, in_path, n_particles):
    distr = [None] * size
    with h5py.File(in_path, 'r', driver=None) as in_file:
        for rank in range(size):
            rank_range, molecules_flag = distribute_input(
                in_file, rank, size, n_particles
            )
            distr[rank] = (rank_range, molecules_flag)
    return distr


def test_file_io_distribute_input_size_1(h5py_molecules_file):
    in_path, n_particles, _, _ = h5py_molecules_file
    size = 1
    distr = _distribute_nranks(size, in_path, n_particles)
    rank_range, molecules_flag = distr[0]
    assert molecules_flag
    assert rank_range[0] == 0
    assert rank_range[-1] == n_particles - 1
    assert len(rank_range) == n_particles


def test_file_io_distribute_input_size_2(h5py_molecules_file):
    in_path, n_particles, indices, molecules = h5py_molecules_file
    size = 2
    distr = _distribute_nranks(size, in_path, n_particles)
    rank_range_0, molecules_flag_0 = distr[0]
    rank_range_1, molecules_flag_1 = distr[1]

    assert molecules_flag_0
    assert molecules_flag_1

    assert rank_range_0[0] == 0
    assert molecules[rank_range_0[-1]] != molecules[rank_range_0[-1] + 1]

    assert molecules[rank_range_1[0] - 1] != molecules[rank_range_1[0]]
    assert rank_range_1[-1] == n_particles - 1

    assert rank_range_1[0] == rank_range_0[-1] + 1
    rank_range_all = np.concatenate((rank_range_0, rank_range_1))
    assert np.equal(rank_range_all, np.arange(n_particles)).all()


def test_file_io_distribute_input_size_3(h5py_molecules_file):
    in_path, n_particles, indices, molecules = h5py_molecules_file
    size = 3
    distr = _distribute_nranks(size, in_path, n_particles)
    rank_range_0, molecules_flag_0 = distr[0]
    rank_range_1, molecules_flag_1 = distr[1]
    rank_range_2, molecules_flag_2 = distr[2]

    assert molecules_flag_0
    assert molecules_flag_1
    assert molecules_flag_2

    assert rank_range_0[0] == 0
    assert molecules[rank_range_0[-1]] != molecules[rank_range_0[-1] + 1]

    assert molecules[rank_range_1[0] - 1] != molecules[rank_range_1[0]]
    assert molecules[rank_range_1[-1]] != molecules[rank_range_1[-1] + 1]

    assert molecules[rank_range_2[0] - 1] != molecules[rank_range_2[0]]
    assert rank_range_2[-1] == n_particles - 1

    rank_range_all = np.concatenate((rank_range_0, rank_range_1, rank_range_2))
    assert np.equal(rank_range_all, np.arange(n_particles)).all()


def test_file_io_distribute_input_various_sizes(h5py_molecules_file):
    in_path, n_particles, indices, molecules = h5py_molecules_file
    for size in (5, 9, 11, 14, 19, 25):
        distr = _distribute_nranks(size, in_path, n_particles)

        assert distr[0][0][0] == 0
        assert distr[-1][0][-1] == n_particles - 1

        for i, d in enumerate(distr):
            molecules_flag = d[1]
            assert molecules_flag

            rank_range = d[0]
            if i == 0:
                assert rank_range[0] == 0
            else:
                assert molecules[rank_range[0] - 1] != molecules[rank_range[0]]

            if i == size - 1:
                assert rank_range[-1] == n_particles - 1
            else:
                assert (molecules[rank_range[-1]] !=
                        molecules[rank_range[-1] + 1])

        rank_range_all = np.concatenate([d[0] for d in distr])
        assert np.equal(rank_range_all, np.arange(n_particles)).all()



import pytest
import numpy as np
import logging
import filecmp
import os
import sys
from mpi4py import MPI
from hymd.input_parser import Config
from hymd.file_io import distribute_input

@pytest.mark.mpi(min_size=2)
def test_plumed_bias_obj(molecules_with_solvent, change_test_dir, tmp_path, 
                         caplog, monkeypatch):
    pytest.importorskip("plumed")
    from hymd.plumed import PlumedBias
    caplog.set_level(logging.INFO)
    indices, positions, molecules, velocities, bonds, names, types = molecules_with_solvent
    box_size = np.array([10, 10, 10], dtype=np.float64)
    config = Config(n_steps=1, time_step=0.03, box_size=box_size,
                    mesh_size=[5, 5, 5], sigma=0.5, kappa=0.05,
                    n_particles=len(indices), target_temperature=300.0)
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    # create PLUMED input with the correct tmp_path
    plumed_str = """d14: DISTANCE ATOMS=1,4

RESTRAINT ARG=d14 KAPPA=200.0 AT=2.0

DUMPATOMS ATOMS=1-45 FILE={}
PRINT ARG=d14 FILE={}""".format(os.path.join(tmp_path,"testdump.xyz"), 
                                os.path.join(tmp_path,"DIST"))
    with open(os.path.join(tmp_path,"plumed.dat"),"w") as f:
        f.write(plumed_str)

    # Test stub acting like a hdf5 file for distribute_input
    in_file = {'molecules': molecules, 'indices': indices}
    rank_range, molecules_flag = distribute_input(
        in_file, rank, size,  config.n_particles, 6, comm=comm
    )
    positions_ = positions[rank_range]
    molecules_ = molecules[rank_range]
    indices_ = indices[rank_range].astype(np.intc)
    velocities_ = velocities[rank_range]
    bonds_ = bonds[rank_range]
    types_ = types[rank_range]
    names_ = names[rank_range]

    # test working case
    plumed_obj = PlumedBias(
        config, 
        os.path.join(tmp_path,"plumed.dat"),
        os.path.join(tmp_path,"plumed.out"),
        comm=comm,
        verbose=2
    )

    if rank == 0:
        log = caplog.text
        print(log)
        cmp_strings = (
            "Attempting to read PLUMED input from",
            "Using PLUMED_KERNEL="
        )
        assert all([(s in log) for s in cmp_strings])
        caplog.clear()
    assert plumed_obj.api_version > 3

    forces_ = np.zeros_like(positions_)
    charges_ = np.zeros_like(indices_, dtype=np.double)

    needs_energy = plumed_obj.prepare(
        0,
        forces_,
        positions_,
        indices_,
        config,
        charges_
    )

    assert needs_energy == False
    if rank == 0:
        log = caplog.text
        cmp_strings = (
            "Setting PLUMED pointers for step",
        )
        assert all([(s in log) for s in cmp_strings])
        caplog.clear()

    plumed_forces, plumed_bias = plumed_obj.calc(forces_, 0.0)
    if rank == 0:
        log = caplog.text
        cmp_strings = (
            "Calculating PLUMED forces",
            "Done calculating PLUMED forces"
        )
        assert all([(s in log) for s in cmp_strings])
        caplog.clear()

    if rank == 0:
        ref_forces = np.loadtxt("refforces.txt")
        assert plumed_forces == pytest.approx(ref_forces, abs=1e-8)
        assert filecmp.cmp(os.path.join(tmp_path,"DIST"), "refDIST") == True
        assert filecmp.cmp(os.path.join(tmp_path,"testdump.xyz"), "reftestdump.xyz") == True

    assert plumed_bias == pytest.approx(635.5621691482, abs=1e-8)

    # test calc without prepare
    with pytest.raises(RuntimeError) as recorded_error:
        _, _ = plumed_obj.calc(forces_, 0.0)
        if rank == 0:
            log = caplog.text
            assert "without first calling prepare method" in log
    message = str(recorded_error.value)
    assert "without first calling prepare method" in message
    caplog.clear()

    # finalize PlumedBias object
    plumed_obj.finalize()


@pytest.mark.mpi()
@pytest.mark.skip(reason="Currently fails in CI due to environment variable")
def test_fail_plumed_bias_obj(monkeypatch):
    pytest.importorskip("plumed")

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    # create fake config
    box_size = np.array([10, 10, 10], dtype=np.float64)
    config = Config(n_steps=1, time_step=0.03, box_size=box_size,
                    mesh_size=[5, 5, 5], sigma=0.5, kappa=0.05,
                    n_particles=1000, target_temperature=300.0)

    # try to get rid of PLUMED_KERNEL path
    monkeypatch.delenv("PLUMED_KERNEL")
    found_plumed_in_syspath = False
    for libpath in sys.path:
        if "plumed2/" in libpath:
            found_plumed_in_syspath = True
            break

    import hymd.plumed
   
    with pytest.raises(RuntimeError) as recorded_error:
        with hymd.plumed.PlumedBias(
            config, 
            "test.in", 
            "test.out",
            comm=comm,
            verbose=2
        ) as _:
            if rank == 0:
                log = caplog.text
                cmp_strings = (
                    "The PLUMED_KERNEL environment variable is not set",
                )
                assert all([(s in log) for s in cmp_strings])

    if not found_plumed_in_syspath:
        cmp_strings = (
            "HyMD was not able to create a PLUMED object",
            "Maybe there is a problem with your PLUMED_KERNEL?"
        )
        message = str(recorded_error.value)
        assert all([(s in message) for s in cmp_strings])


def test_unavailable_plumed(hide_available_plumed):
    import hymd.plumed

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    # create fake config
    box_size = np.array([10, 10, 10], dtype=np.float64)
    config = Config(n_steps=1, time_step=0.03, box_size=box_size,
                    mesh_size=[5, 5, 5], sigma=0.5, kappa=0.05,
                    n_particles=1000, target_temperature=300.0)

    with pytest.raises(ImportError) as recorded_error:
        with hymd.plumed.PlumedBias(
            config, 
            "test.in", 
            "test.out",
            comm=comm,
            verbose=2
        ) as _:
            if rank == 0:
                log = caplog.text
                cmp_strings = (
                    "You are trying to use PLUMED",
                    "but HyMD could not import py-plumed."
                )
                assert all([(s in log) for s in cmp_strings])
    cmp_strings = (
        "You are trying to use PLUMED",
        "but HyMD could not import py-plumed."
    )
    message = str(recorded_error.value)
    assert all([(s in message) for s in cmp_strings])



"""Forces and energies from the discretized particle-field grid interactions
"""
import logging
import numpy as np
from mpi4py import MPI
from .logger import Logger
import warnings


def initialize_pm(pmesh, config, comm=MPI.COMM_WORLD):
    """
    Creates the necessary pmesh objects for pfft operations.

    Parameters
    ----------
    pmesh : module 'pmesh.pm'
    config : Config
        Configuration dataclass containing simulation metadata and parameters.
    comm : MPI.Intracomm, optional
        MPI communicator to use for rank commuication. Defaults to
        MPI.COMM_WORLD.

    Returns
    -------
    pm : object 'pmesh.pm.ParticleMesh'
    field_list : list[pmesh.pm.RealField], (multiple)
        Essential list of pmesh objects required for MD
    list_coulomb : list[pmesh.pm.RealField], (multiple)
        Additional list of pmesh objects required for electrostatics.
    """

    if config.dtype == np.float64:
        pmeshtype = "f8"
    else:
        pmeshtype = "f4"
    # Ignore numpy numpy.VisibleDeprecationWarning: Creating an ndarray from
    # ragged nested sequences until it is fixed in pmesh
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            category=np.VisibleDeprecationWarning,
            message=r"Creating an ndarray from ragged nested sequences",
        )
        # The first argument of ParticleMesh has to be a tuple
        pm = pmesh.ParticleMesh(
            config.mesh_size, BoxSize=config.box_size, dtype=pmeshtype, comm=comm
        )
    phi = [pm.create("real", value=0.0) for _ in range(config.n_types)]
    phi_fourier = [
        pm.create("complex", value=0.0) for _ in range(config.n_types)
    ]  # noqa: E501
    force_on_grid = [
        [pm.create("real", value=0.0) for d in range(3)] for _ in range(config.n_types)
    ]
    v_ext_fourier = [pm.create("complex", value=0.0) for _ in range(4)]
    v_ext = [pm.create("real", value=0.0) for _ in range(config.n_types)]

    phi_transfer = [pm.create("complex", value=0.0) for _ in range(3)]

    phi_laplacian = [
        [pm.create("real", value=0.0) for d in range(3)] for _ in range(config.n_types)
    ]

    # Initialize charge density fields
    coulomb_list = []
    elec_common_list = [None, None, None, None]
    _SPACE_DIM = config.box_size.size

    if config.coulombtype == "PIC_Spectral_GPE" or config.coulombtype == "PIC_Spectral":
        phi_q = pm.create("real", value=0.0)
        phi_q_fourier = pm.create("complex", value=0.0)
        psi = pm.create("real", value=0.0)
        elec_field = [pm.create("real", value=0.0) for _ in range(_SPACE_DIM)]

        elec_common_list = [phi_q, phi_q_fourier, psi, elec_field]

    if config.coulombtype == "PIC_Spectral":
        elec_field_fourier = [
            pm.create("complex", value=0.0) for _ in range(_SPACE_DIM)
        ]  # for force calculation
        psi_fourier = pm.create("complex", value=0.0)

        coulomb_list = [
            elec_field_fourier,
            psi_fourier,
        ]

    if (
        config.coulombtype == "PIC_Spectral_GPE"
    ):  ## initializing the density mesh #dielectric_flag
        phi_eps = pm.create(
            "real", value=0.0
        )  ## real contrib of the epsilon dielectric painted to grid
        phi_eps_fourier = pm.create("complex", value=0.0)  # complex contrib of phi eps
        phi_eta = [
            pm.create("real", value=0.0) for _ in range(_SPACE_DIM)
        ]  ## real contrib of factor in polarization charge density
        phi_eta_fourier = [
            pm.create("complex", value=0.0) for _ in range(_SPACE_DIM)
        ]  ## fourier of factor in polarization charge density
        phi_pol = pm.create(
            "real", value=0.0
        )  ## real contrib of the polarization charge
        phi_pol_prev = pm.create("real", value=0.0)
        elec_dot = pm.create("real", value=0.0)
        elec_field_contrib = pm.create(
            "real", value=0.0
        )  # needed for pol energies later

        # External potential and force meshes
        Vbar_elec = [pm.create("real", value=0.0) for _ in range(config.n_types)]
        Vbar_elec_fourier = [
            pm.create("complex", value=0.0) for _ in range(config.n_types)
        ]
        force_mesh_elec = [
            [pm.create("real", value=0.0) for d in range(3)]
            for _ in range(config.n_types)
        ]
        force_mesh_elec_fourier = [
            [pm.create("complex", value=0.0) for d in range(3)]
            for _ in range(config.n_types)
        ]

        coulomb_list = [
            phi_eps,
            phi_eps_fourier,
            phi_eta,
            phi_eta_fourier,
            phi_pol,
            phi_pol_prev,
            elec_dot,
            elec_field_contrib,
            Vbar_elec,
            Vbar_elec_fourier,
            force_mesh_elec,
            force_mesh_elec_fourier,
        ]

    field_list = [
        phi,
        phi_fourier,
        force_on_grid,
        v_ext_fourier,
        v_ext,
        phi_transfer,
        phi_laplacian,
    ]

    return (pm, field_list, elec_common_list, coulomb_list)


def compute_field_force(layouts, r, force_mesh, force, types, n_types):
    """Interpolate particle-field forces from the grid onto particle positions

    Backmaps the forces calculated on the grid to particle positions using the
    window function :math:`P` (by default cloud-in-cell [CIC]). In the
    following, let :math:`\\mathbf{F}_j` denote the force acting on the
    particle with index :math:`j` and position :math:`\\mathbf{r}_j`. The
    interpolated force is

    .. math::

        \\mathbf{F}_j = -\\sum_k\\nabla V_{j_k}
            P(\\mathbf{r}_{j_k}-\\mathbf{r}_j)h^3,

    where :math:`V_{j_k}` is the discretized external potential at grid vertex
    :math:`j_k`, :math:`\\mathbf{r}_{j_k}` is the position of the grid vertex
    with index :math:`j_k`, and :math:`h^3` is the volume of each grid voxel.
    The sum is taken over all closest neighbour vertices, :math:`j_k`.

    Parameters
    ----------
    layouts : list[pmesh.domain.Layout]
        Pmesh communication layout objects for domain decompositions of each
        particle type. Used as blueprint by :code:`pmesh.pm.readout` for
        exchange of particle information across MPI ranks as necessary.
    r : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    force_mesh : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        force density values on the computational grid; :code:`D` fields in D
        dimensions for each particle type. Local for each MPI rank--the full
        computational grid is represented by the collective fields of all MPI
        ranks.
    force : (N,D) numpy.ndarray
        Array of forces for :code:`N` particles in :code:`D` dimensions. Local
        for each MPI rank.
    types : (N,) numpy.ndarray
        Array of type indices for each of :code:`N` particles. Local for each
        MPI rank.
    n_types : int
        Number of different unique types present in the simulation system.
        :code:`n_types` is global, i.e. the same for all MPI ranks even if some
        ranks contain zero particles of certain types.
    """
    for t in range(n_types):
        ind = types == t
        for d in range(3):
            force[ind, d] = force_mesh[t][d].readout(r[ind], layout=layouts[t])


def compute_self_energy_q(config, charges, comm=MPI.COMM_WORLD):
    """Compute the self energy for the interaction due to the Ewald scheme
    used to compute the electrostatics. The energy is stored in :code:`config`.

    The self interaction energy is given by:

    .. math::

        U_{self} = \\sqrt{\\frac{1}{2\\pi\\sigma^2}} \\sum_{i=1}^{N} q_i^2

    where :math:`q_i` are the charges and :math:`\\sigma` is the half-width
    of the Gaussian filter.

    Parameters
    ----------
    config : Config
        Configuration object.
    charges : (N,) numpy.ndarray
        Array of particle charge values for :code:`N` particles. Local for each
        MPI rank.
    comm : mpi4py.Comm
        MPI communicator to use for rank communication.

    Returns
    -------
    field_q_self_energy : float
        Electrostatic self energy.
    """
    elec_conversion_factor = config.coulomb_constant / config.dielectric_const

    prefac = elec_conversion_factor * np.sqrt(
        1.0 / (2.0 * np.pi * config.sigma * config.sigma)
    )
    _squared_charges = charges * charges
    squared_charges_sum = comm.allreduce(np.sum(_squared_charges))
    return prefac * squared_charges_sum


def update_field_force_q(
    charges,
    phi_q,
    phi_q_fourier,
    psi,
    psi_fourier,
    elec_field_fourier,
    elec_field,
    elec_forces,
    layout_q,
    hamiltonian,
    pm,
    positions,
    config,
):
    """Calculate the electrostatic particle-field forces on the grid

    Computes the electrostatic potential :math:`\\Psi` from particle charges
    through the smoothed charge density :math:`\\tilde\\rho`. With :math:`P`
    being the cloud-in-cell (CIC) window function, the charge density and
    filtered charge densities are computed as

    .. math::

        \\rho(\\mathbf{r}) = \\sum_i q_i P(\\mathbf{r}-\\mathbf{r}_i),

    and

    .. math::

        \\tilde\\rho(\\mathbf{r}) = \\int\\mathrm{x}\\mathbf{r}\\,
            \\rho(\\mathbf{x})H(\\mathbf{r}-\\mathbf{x}),

    where :math:`H` is the grid-independent filtering function. The
    electrostatic potential is computed in reciprocal space as

    .. math::

        \\Phi = \\mathrm{FFT}^{-1}\\left[
            \\frac{4\\pi k_e}{\\varepsilon \\vert\\mathbf{k}\\vert^2}
            \\mathrm{FFT}(\\rho)\\mathrm{FFT}(H)
        \\right],

    with the electric field

    .. math::

        \\mathbf{E} = \\mathrm{FFT}^{-1}\\left[
            -i\\mathbf{k}\\,\\mathrm{FFT}(\\Psi)
        \\right].

    In the following, let :math:`\\mathbf{F}_j` denote the electrostatic force
    acting on the particle with index :math:`j` and position
    :math:`\\mathbf{r}_j`. The interpolated electrostatic force is

    .. math::

        \\mathbf{F}_j = \\sum_k q_j\\mathbf{E}_{j_k}
            P(\\mathbf{r}_{j_k}-\\mathbf{r}_j)h^3,

    where :math:`\\mathbf{E}_{j_k}` is the discretized electric field at grid
    vertex :math:`j_k`, :math:`\\mathbf{r}_{j_k}` is the position of the grid
    vertex with index :math:`j_k`, and :math:`h^3` is the volume of each grid
    voxel. The sum is taken over all closest neighbour vertices, :math:`j_k`.

    Parameters
    ----------
    charges : (N,) numpy.ndarray
        Array of particle charge values for :code:`N` particles. Local for each
        MPI rank.
    phi_q : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        charge density density values on the computational grid. Pre-allocated,
        but empty; any values in this field are discarded. Changed in-place.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    phi_q_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        Fourier transformed charge density values in reciprocal space on the
        computational grid. Pre-allocated, but empty; any values in this field
        are discarded. Changed in-place. Local for each MPI rank--the full
        computational grid is represented by the collective fields of all MPI
        ranks.
    elec_field_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        electric field values in reciprocal space on the computational grid.
        Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computational grid
        is represented by the collective fields of all MPI ranks.
    elec_field : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        electric field values on the computational grid. Pre-allocated,
        but empty; any values in this field are discarded. Changed in-place.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    elec_forces : (N,D) numpy.ndarray
        Array of electrostatic forces on :code:`N` particles in :code:`D`
        dimensions.
    layout_q : pmesh.domain.Layout
        Pmesh communication layout object for domain decomposition of the full
        system. Used as blueprint by :code:`pmesh.pm.paint` and
        :code:`pmesh.pm.readout` for exchange of particle information across
        MPI ranks as necessary.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    config : hymd.input_parser.Config
        Configuration object.
    """
    V = np.prod(config.box_size)
    n_mesh_cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh_cells
    n_dimensions = config.box_size.size
    elec_conversion_factor = config.coulomb_constant / config.dielectric_const

    # charges to grid
    pm.paint(positions, layout=layout_q, mass=charges, out=phi_q)
    phi_q /= volume_per_cell
    # print("phi_q", np.sum(phi_q))
    phi_q.r2c(out=phi_q_fourier)

    # smear charges with filter
    phi_q_fourier.apply(hamiltonian.H, out=phi_q_fourier)

    # solve Poisson equation in Fourier space to get electrostatic potential
    def poisson_transfer_function(k, v):
        return 4.0 * np.pi * elec_conversion_factor * v / k.normp(p=2, zeromode=1)

    phi_q_fourier.apply(poisson_transfer_function, out=psi_fourier)
    # print("psi_fourier", np.sum(psi_fourier))
    psi_fourier.c2r(out=psi)
    # print("phi_q * psi update_field", np.sum(phi_q * psi))

    # exit()

    # compute electric field directly from smeared charged densities in Fourier
    for _d in np.arange(n_dimensions):

        def poisson_transfer_function(k, v, d=_d):
            return (
                -1j
                * k[d]
                * 4.0
                * np.pi
                * elec_conversion_factor
                * v
                / k.normp(p=2, zeromode=1)
            )

        phi_q_fourier.apply(poisson_transfer_function, out=elec_field_fourier[_d])
        elec_field_fourier[_d].c2r(out=elec_field[_d])

    # get electrostatic force from electric field
    for _d in np.arange(n_dimensions):
        elec_forces[:, _d] = charges * (
            elec_field[_d].readout(positions, layout=layout_q)
        )


def comp_laplacian(
    phi_fourier,
    phi_transfer,
    phi_laplacian,
    hamiltonian,
    config,
):
    for t in range(config.n_types):
        np.copyto(phi_transfer[0].value, phi_fourier[t].value, casting="no", where=True)
        np.copyto(phi_transfer[1].value, phi_fourier[t].value, casting="no", where=True)
        np.copyto(phi_transfer[2].value, phi_fourier[t].value, casting="no", where=True)

        # Evaluate laplacian of phi in fourier space
        for d in range(3):

            def laplacian_transfer(k, v, d=d):
                return -k[d] ** 2 * v

            phi_transfer[d].apply(laplacian_transfer, out=Ellipsis)
            phi_transfer[d].c2r(out=phi_laplacian[t][d])


def update_field(
    phi,
    phi_laplacian,
    phi_transfer,
    layouts,
    force_mesh,
    hamiltonian,
    pm,
    positions,
    types,
    config,
    v_ext,
    phi_fourier,
    v_ext_fourier,
    m,
    compute_potential=False,
):
    """Calculate the particle-field potential and force density

    If :code:`compute_potential` is :code:`True`, the energy may subsequently
    be computed by calling :code:`compute_field_and_kinetic_energy`.

    Computes the particle-field external potential :math:`V_\\text{ext}` from
    particle number densities through the smoothed density field,
    :math:`\\phi(\\mathbf{r})`. With :math:`P` being the cloud-in-cell (CIC)
    window function, the density and filtered densities are computed as

    .. math::

        \\phi(\\mathbf{r}) = \\sum_i P(\\mathbf{r}-\\mathbf{r}_i),

    and

    .. math::

        \\tilde\\phi(\\mathbf{r}) = \\int\\mathrm{x}\\mathbf{r}\\,
            \\phi(\\mathbf{x})H(\\mathbf{r}-\\mathbf{x}),

    where :math:`H` is the grid-independent filtering function. The
    external potential is computed in reciprocal space as

    .. math::

        V_\\text{ext} = \\mathrm{FFT}^{-1}\\left[
            \\mathrm{FFT}
                \\left(
                    \\frac{\\partial w}{\\partial \\tilde\\phi}
                \\right)
            \\mathrm{FFT}(H)
        \\right],

    where :math:`w` is the interaction energy functional. Differentiating
    :math:`V_\\text{ext}` is done by simply applying :math:`i\\mathbf{k}` in
    Fourier space, and the resulting forces are back-transformed to direct
    space and interpolated to particle positions by

    .. math::

        \\mathbf{F}_j = -\\sum_{j_k} \\nabla V_{j_k}
            P(\\mathbf{r}_{j_k} - \\mathbf{r}_j)h^3,

    where :math:`j_k` are the neighbouring vertices of particle :math:`j` at
    position :math:`\\mathbf{r}_j`, and :math:`h^3` is the volume of each grid
    voxel.

    Parameters
    ----------
    phi : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type.
        Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computational grid
        is represented by the collective fields of all MPI ranks.
    phi_laplacian : list[pmesh.pm.RealField], (M, 3)
        Like phi, but containing the laplacian of particle number densities.
    phi_transfer : list[pmesh.pm.ComplexField], (3,)
        Like phi_fourier, used as an intermediary to perform FFT operations
        to obtain the gradient or laplacian of particle number densities.
    layouts : list[pmesh.domain.Layout]
        Pmesh communication layout objects for domain decompositions of each
        particle type. Used as blueprint by :code:`pmesh.pm.readout` for
        exchange of particle information across MPI ranks as necessary.
    force_mesh : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        force density values on the computational grid; :code:`D` fields in D
        dimensions for each particle type. Pre-allocated, but empty; any values
        in this field are discarded. Changed in-place. Local for each MPI
        rank--the full computational grid is represented by the collective
        fields of all MPI ranks.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    types : (N,) numpy.ndarray
        Array of type indices for each of :code:`N` particles. Local for each
        MPI rank.
    config : Config
        Configuration object.
    v_ext : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        external potential values on the computational grid; one for each
        particle type. Pre-allocated, but empty; any values in this field are
        discarded Changed in-place. Local for each MPI rank--the full
        computational grid is represented by the collective fields of all MPI
        ranks.
    phi_fourier : list[pmesh.pm.ComplexField]
        Pmesh :code:`ComplexField` objects containing discretized particle
        number density values in reciprocal space on the computational grid;
        one for each particle type. Pre-allocated, but empty; any values in
        this field are discarded Changed in-place. Local for each MPI rank--the
        full computational grid is represented by the collective fields of all
        MPI ranks.
    v_ext_fourier : list[pmesh.pm.ComplexField]
        Pmesh :code:`ComplesField` objects containing discretized
        particle-field external potential values in reciprocal space on the
        computational grid; :code:`D+1` fields in D dimensions for each
        particle type. :code:`D` copies are made after calculation for later
        use in force calculation, because the `force transfer function`
        application differentiates the field in-place, ruining the contents
        for differentiation in the remaining :code:`D-1` spatial directions.
        Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computational grid
        is represented by the collective fields of all MPI ranks.
    m: list[float], (M,)
        pmesh.pm.ParticleMesh parameter for mass of particles in simulation unit.
        Defaults to 1.0 for all particle types.
    compute_potential : bool, optional
        If :code:`True`, a :code:`D+1`-th copy of the Fourier transformed
        external potential field is made to be used later in particle-field
        energy calculation. If :code:`False`, only :code:`D` copies are made.

    See also
    --------
    compute_field_and_kinetic_energy :
        Compute the particle-field energy after the external potential is
        calculated.
    """
    V = np.prod(config.box_size)
    n_mesh_cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh_cells
    for t in range(config.n_types):
        pm.paint(positions[types == t], mass=m[t], layout=layouts[t], out=phi[t])
        phi[t] /= volume_per_cell
        phi[t].r2c(out=phi_fourier[t])
        phi_fourier[t].apply(hamiltonian.H, out=Ellipsis)
        phi_fourier[t].c2r(out=phi[t])

    # External potential
    for t in range(config.n_types):
        v = hamiltonian.v_ext[t](phi)

        v.r2c(out=v_ext_fourier[0])
        v_ext_fourier[0].apply(hamiltonian.H, out=Ellipsis)
        np.copyto(
            v_ext_fourier[1].value,
            v_ext_fourier[0].value,
            casting="no",
            where=True,
        )
        np.copyto(
            v_ext_fourier[2].value,
            v_ext_fourier[0].value,
            casting="no",
            where=True,
        )
        if compute_potential:
            np.copyto(
                v_ext_fourier[3].value,
                v_ext_fourier[0].value,
                casting="no",
                where=True,
            )

        # Differentiate the external potential in fourier space
        for d in range(3):

            def force_transfer_function(k, v, d=d):
                return -k[d] * 1j * v

            v_ext_fourier[d].apply(force_transfer_function, out=Ellipsis)
            v_ext_fourier[d].c2r(out=force_mesh[t][d])

        if compute_potential:
            v_ext_fourier[3].c2r(out=v_ext[t])


def compute_field_and_kinetic_energy(
    phi,
    phi_q,
    psi,
    velocity,
    hamiltonian,
    positions,
    types,
    v_ext,
    config,
    layouts,
    comm=MPI.COMM_WORLD,
):
    """Compute the particle-field and kinetic energy contributions

    Calculates the kinetic energy through

    .. math::

        E_k = \\sum_j \\frac{1}{2}m_j\\mathbf{v}_j\\cdot\\mathbf{v}_j,

    and the particle-field energy by

    .. math::

        E_\\text{field} = \\int\\mathrm{d}\\mathbf{r}\\,
            w[\\tilde\\phi],

    where :math:`w` is the interaction energy functional `density`.

    Parameters
    ----------
    phi : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    velocity : (N,D) numpy.ndarray
        Array of velocities for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    types : (N,) numpy.ndarray
        Array of type indices for each of :code:`N` particles. Local for each
        MPI rank.
    v_ext : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        external potential values on the computational grid; one for each
        particle type. Local for each MPI rank--the full computational grid is
        represented by the collective fields of all MPI ranks.
    config : Config
        Configuration object.
    layouts : list[pmesh.domain.Layout]
        Pmesh communication layout objects for domain decompositions of each
        particle type. Used as blueprint by :code:`pmesh.pm.readout` for
        exchange of particle information across MPI ranks as necessary.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    See also
    --------
    update_field :
        Computes the up-to-date external potential for use in calculating the
        particle-field energy.
    """
    V = np.prod(config.box_size)
    n_mesh__cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh__cells

    w_0 = hamiltonian.w_0(phi) * volume_per_cell
    field_energy = w_0.csum()  # w to W

    kinetic_energy = comm.allreduce(0.5 * config.mass * np.sum(velocity**2))

    if config.coulombtype == "PIC_Spectral":
        w_elec = hamiltonian.w_elec([phi_q, psi]) * volume_per_cell
        field_q_energy = w_elec.csum()
    else:
        field_q_energy = 0.0

    return field_energy, kinetic_energy, field_q_energy


def compute_field_energy_q_GPE(
    config,
    phi_eps,
    field_q_energy,
    dot_elec,
    comm=MPI.COMM_WORLD,
):
    """
    Compute the electrostatic energy after electrosatic forces is
    calculated.

    From the definition of the elecrostatic potential :math:`\\Psi`, the energy
    is

        W = \\frac{1}{2}\\int\\mathrm{d}\\mathbf{r}\\,
            \\epsilon(\\mathbf{r})} \\left(\\mathbf{E}\\cdot \\mathbf{E}\\right),

    where :math:`\\epsilon(\\mathbf{r})}` is the anisotropic, spatially dependent,
    relative dielectric of the simulation medium.

    Parameters
    ----------
    config : hymd.input_parser.Config
        Configuration object.
    phi_eps : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        relative dielectric values on the computational grid.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    field_q_energy : float
        Total elecrostatic energy.
    dot_elec : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing :math:`|\\mathbf{E(r)}|^{2}`
        on the computational grid. Local for each MPI rank -- the full computational
        grid is represented by the collective fields of all MPI ranks.
    comm : mpi4py.Comm
        MPI communicator to use  for rank commuication.

    See also
    --------
    update_field_force_q_GPE:
        Compute the electrosatic force from an anisotropic dielectric general
        Poisson equation.
    """

    V = np.prod(config.box_size)
    n_mesh__cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh__cells
    # ^ due to integration on local cell before allreduce

    eps_0 = 1.0 / (config.coulomb_constant * 4 * np.pi)
    field_q_energy = (
        volume_per_cell * (0.5 * eps_0) * comm.allreduce(np.sum(phi_eps * dot_elec))
    )

    return field_q_energy


def update_field_force_q_GPE(
    conv_fun,
    phi,
    types,
    charges,
    phi_q,
    phi_q_fourier,
    phi_eps,
    phi_eps_fourier,
    phi_eta,
    phi_eta_fourier,
    phi_pol_prev,
    phi_pol,
    elec_field,
    elec_forces,
    elec_field_contrib,
    psi,
    Vbar_elec,
    Vbar_elec_fourier,
    force_mesh_elec,
    force_mesh_elec_fourier,
    hamiltonian,
    layout_q,
    layouts,
    pm,
    positions,
    config,
    comm=MPI.COMM_WORLD,
):
    """
    Calculate the electrostatic particle-field forces on the grid, arising from
    a general Poisson equation, i.e. anisotropic permittivity/dielectric.
    The function is called when tomli input config.coulombtype = "PIC_Spectral_GPE."

    Computes the electrostatic potential :math:`\\Psi` from particle charges
    through the smoothed charge density :math:`\\tilde\\rho`. With :math:`P`
    being the cloud-in-cell (CIC) window function, the charge density and
    filtered charge densities are computed as

    .. math::

        \\rho(\\mathbf{r}) = \\sum_i q_i P(\\mathbf{r}-\\mathbf{r}_i),

    and

    .. math::

        \\tilde\\rho(\\mathbf{r}) = \\int\\mathrm{x}\\mathbf{r}\\,
            \\rho(\\mathbf{x})H(\\mathbf{r}-\\mathbf{x}),

    where :math:`H` is the grid-independent filtering function. The
    electrostatic potential for a variable dielectric does not have an
    analytical expression, and is computed in reciprocal through an iterative
    method.

    The GPE states that

    .. math::

            \\nabla \\cdot \\left(\\epsilon(\\mathbf{r})
            \\nabla{\\mathbf{\\psi(r)}}\\right) = -\\rho({\\mathbf{r}}).

    where :math:`\\epsilon(\\mathbf{r})` is the relative dielectric function.

    Parameters
    ----------
    conv_fun : Convergence function.
        Returns a scalar. Depends on MPI allreduce for similar convergence
        across MPI ranks.
    phi : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    types : (N,) numpy.ndarray
        Array of type indices for each of :code:`N` particles. Local for each
        MPI rank.
    charges : (N,) numpy.ndarray
        Array of particle charge values for :code:`N` particles. Local for each
        MPI rank.
    phi_q : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        charge density density values on the computational grid. Pre-allocated,
        but empty. Changed in-place.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    phi_q_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        Fourier transformed charge density values in reciprocal space on the
        computational grid. Pre-allocated, but empty. Changed in-place.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    phi_eps : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        relative dielectric values on the computational grid. Pre-allocated,
        but empty. Changed in-place.
        Local for each MPI rank--the full computational grid is represented by
        the collective fields of all MPI ranks.
    phi_eps_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        Fourier transformed relative dielectric values in reciprocal space on the
        computational grid. Pre-allocated, but empty. Changed in-place.
        Local for each MPI rank--the full computational grid is represented
        by the collective fields of all MPI ranks.
    phi_eta : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        gradients of the relative dielectric values on the computational grid.
        Pre-allocated,but empty. Changed in-place.Local for each MPI rank--the
        full computational grid is represented by the collective fields of all MPI ranks.
    phi_eta_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing the calculated discretized
        Fourier transformed gradient relative dielectric values in reciprocal space on the
        computational grid. Pre-allocated, but empty.  Changed in-place.
        Local for each MPI rank--the full computational grid is represented
        by the collective fields of all MPI ranks.
    phi_pol_prev : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        polarization charge values on the computational grid. Parameter in
        the iterative method.Pre-allocated,but empty. Changedin-place.
        Local for each MPI rank--the full computational grid is represented
        by the collective fields of all MPI ranks.
    phi_pol : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        polarization charges on the computational grid. Parameter in the iterative
        method, updating the next quess in solving for the electrostatic potential.
        Pre-allocated,but empty.  Changed in-place.Local for each MPI rank--
        the full computational grid is represented by the collective fields of
        all MPI ranks.
    elec_field : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        electric field values on the computational grid. Pre-allocated,
        but empty. Changed in-place. Local for each MPI rank--the full
        computational grid is represented by the collective fields of all
        MPI ranks.
    elec_forces : (N,D) numpy.ndarray
        Array of electrostatic forces on :code:`N` particles in :code:`D`
        dimensions.
    elec_field_contrib : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing
        :math:`|\\mathbf{E(r)}|^2/\\phi_{0}` on the computational grid.
        Pre-allocated, but empty. Changed in-place.
        Local for each MPI rank-- the full computational grid is represented by
        the collective fields of all MPI ranks.
    psi : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing electrostatic potential
        on the computational grid. Pre-allocated, but empty. Changed in-place.
        Local for each MPI rank-- the full computational grid is represented by
        the collective fields of all MPI ranks.
    Vbar_elec : mesh.pm.RealField
        Pmesh :code:`RealField` object for storing functional derivatives of
        :math:`\\|w(\\{ \\phi \\})_{elec}`on the computational grid.
        Pre-allocated, but empty. Changed in-place. Local for each MPI rank--
        the full computational grid is represented by the collective fields of
         all MPI ranks.
    Vbar_elec_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing the calculated functional
        derivatives of :math:`\\|w(\\{ \\phi \\})_{elec}` in reciprocal space on the
        computational grid. Pre-allocated, but empty.  Changed in-place.
        Local for each MPI rank--the full computational grid is represented
        by the collective fields of all MPI ranks.
    force_mesh_elec : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing electrostatic force values
        on the computational grid. Pre-allocated, but empty. Changed in-place.
        Local for each MPI rank-- the full computational grid is represented by
        the collective fields of all MPI ranks.
    force_mesh_elec_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing the calculated electrostatic
        force values in reciprocal space on the computational grid. Local for
        each MPI rank--the full computational grid is represented by the collective
        fields of all MPI ranks.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    layout_q : pmesh.domain.Layout
        Pmesh communication layout object for domain decomposition of the full
        system. Used as blueprint by :code:`pmesh.pm.paint` and
        :code:`pmesh.pm.readout` for exchange of particle information across
        MPI ranks as necessary.
    layouts: list[pmesh.domain.Layout]
        Pmesh communication layout objects for domain decompositions of each
        particle type. Used as blueprint by :code:`pmesh.pm.readout` for
        exchange of particle information across MPI ranks as necessary.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    config : hymd.input_parser.Config
        Configuration object.
    comm: mpi4py.Comm
        MPI communicator to use for rank commuication.

    See also
    --------
    compute_field_energy_q_GPE:
        Compute the electrostatic energy after electrosatic force is
        calculated for a variable (anisotropic) dielectric general Poisson equation.
    """

    ## basic setup
    V = np.prod(config.box_size)
    n_mesh_cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh_cells
    ## old protocol in gen_qe_hpf_use_self
    pm.paint(positions, layout=layout_q, mass=charges, out=phi_q)  ##
    ## scale and fft
    ## old protocol in gen_qe_hpf_use_self
    phi_q /= volume_per_cell
    phi_q.r2c(out=phi_q_fourier)

    phi_q_fourier.apply(hamiltonian.H, out=phi_q_fourier)
    ## ^------ use the same gaussian as the \kai interaciton
    phi_q_fourier.c2r(out=phi_q)  ## this phi_q is after applying the smearing function

    denom_phi_tot = pm.create("real", value=0.0)
    num_types = pm.create("real", value=0.0)
    ### ^ ----- Calculate the relative dielectric (permittivity) to field
    ### ------- from a mean contribution of particle number densities

    for t_ in range(config.n_types):
        num_types = num_types + (config.dielectric_type[t_]) * phi[t_]
        denom_phi_tot = denom_phi_tot + phi[t_]

    np.divide(num_types, denom_phi_tot, where=np.abs(denom_phi_tot > 1e-6), out=phi_eps)

    phi_eps.r2c(out=phi_eps_fourier)  # FFT dielectric

    # phi_q_eps = (phi_q/phi_eps)
    np.divide(phi_q, phi_eps, where=np.abs(phi_eps > 1e-6), out=phi_q)

    _SPACE_DIM = 3
    ##^--------- constants needed throughout the calculations

    ### method for finding the gradient (fourier space), using the spatial dimension of k
    for _d in np.arange(_SPACE_DIM):

        def gradient_transfer_function(k, x, d=_d):
            return 1j * k[d] * x

        phi_eps_fourier.apply(gradient_transfer_function, out=phi_eta_fourier[_d])
        phi_eta_fourier[_d].c2r(out=phi_eta[_d])
        np.divide(phi_eta[_d], phi_eps, where=np.abs(phi_eps > 1e-6), out=phi_eta[_d])

    ### iterative GPE solver ###
    ### ----------------------------------------------
    max_iter = 100
    i = 0
    delta = 1.0
    # phi_pol_prev = pm.create("real", value = 0.0)
    ### ^------ set to zero before each iterative procedure or soft start
    conv_criteria = config.conv_crit  # conv. criteria (default 1e-6)
    w = config.pol_mixing  # polarization mixing param (default 0.6)
    while i < max_iter and delta > conv_criteria:
        (phi_q + phi_pol_prev).r2c(out=phi_q_fourier)
        for _d in np.arange(_SPACE_DIM):

            def iterate_apply_k_vec(k, additive_terms, d=_d):
                return additive_terms * (-1j * k[d]) / k.normp(p=2, zeromode=1)

            phi_q_fourier.apply(iterate_apply_k_vec, out=phi_eta_fourier[_d])
            phi_eta_fourier[_d].c2r(out=elec_field[_d])

        phi_pol = -(
            phi_eta[0] * elec_field[0]
            + phi_eta[1] * elec_field[1]
            + phi_eta[2] * elec_field[2]
        )
        ### ^-- Following a negative sign convention (-ik) of the FT, a neg sign is
        ### --- mathematically correct by the definition of the GPE (double  - -> +)
        phi_pol = w * phi_pol + (1.0 - w) * phi_pol_prev
        diff = np.abs(phi_pol - phi_pol_prev)
        delta = conv_fun(comm, diff)  # decided from toml input
        phi_pol_prev = phi_pol.copy()
        i = i + 1
    # print("Stopping after iteration {:d} with stop crit {:.2e}, delta {:.2e}".format(i,conv_criteria,delta))

    # compute_potential = True
    def k_norm_divide(k, potential):
        return potential / k.normp(p=2, zeromode=1)

    ## > Electrostatic potential
    eps0_inv = config.coulomb_constant * 4 * np.pi
    ## ^ the 1/(4pi eps0)*4*pi = 1/eps0
    ((eps0_inv) * (phi_q + phi_pol)).r2c(out=phi_q_fourier)
    phi_q_fourier.apply(k_norm_divide, out=phi_q_fourier)
    phi_q_fourier.c2r(out=psi)
    ### ^ electrostatic potential for the GPE

    for _d in np.arange(_SPACE_DIM):

        def field_transfer_function(k, x, d=_d):
            return (
                -1j * k[d] * x
            )  ## negative sign relation, due to E = - nabla psi relation

        phi_q_fourier.apply(field_transfer_function, out=phi_eta_fourier[_d])
        phi_eta_fourier[_d].c2r(out=elec_field[_d])
    ## ^-------- Method: Obtaining the electric field from electrostatic potential
    ## Assuming the electric field is conserved.
    ## Assumption holds if no magnetic flux (magnetic induced fields)

    ##############  Obtain forces  ##############
    elec_dot = (
        elec_field[0] * elec_field[0]
        + elec_field[1] * elec_field[1]
        + elec_field[2] * elec_field[2]
    )
    # needed for energy calculations

    np.divide(
        elec_dot,
        denom_phi_tot,
        where=np.abs(denom_phi_tot > 1e-6),
        out=elec_field_contrib,
    )

    eps0_inv = config.coulomb_constant * 4 * np.pi

    for t_ in range(config.n_types):
        Vbar_elec[t_] = (
            config.type_charges[t_] * psi
            - (0.5 / eps0_inv)
            * (config.dielectric_type[t_] - phi_eps)
            * elec_field_contrib
        )

    # Obtain Vext,k
    for t_ in range(config.n_types):
        Vbar_elec[t_].r2c(out=Vbar_elec_fourier[t_])
        Vbar_elec_fourier[t_].apply(hamiltonian.H, out=Vbar_elec_fourier[t_])

    # force terms
    # F = - grad Vext
    for t_ in range(config.n_types):
        for _d in np.arange(_SPACE_DIM):

            def force_transfer_function(k, x, d=_d):
                return -1j * k[_d] * x  ## negative gradient

            Vbar_elec_fourier[t_].apply(
                force_transfer_function, out=force_mesh_elec_fourier[t_][_d]
            )
            force_mesh_elec_fourier[t_][_d].c2r(out=force_mesh_elec[t_][_d])
            elec_forces[types == t_, _d] = force_mesh_elec[t_][_d].readout(
                positions[types == t_], layout=layouts[t_]
            )

    return Vbar_elec, phi_eps, elec_dot


def domain_decomposition(
    positions, pm, *args, molecules=None, bonds=None, verbose=0, comm=MPI.COMM_WORLD
):
    """Performs domain decomposition

    Rearranges the MPI rank memory layout into one that better mimics the PFFT
    pencil grid 2D decomposition. As molecules are always required to be fully
    contained on a single MPI rank, a perfect decomposition is not always
    possible. The best decomposition which leaves the center of mass of all
    molecules in their respective correct domains is used, with possible
    overlap into neighbouring domains if the spatial extent of any molecule
    crosses domain boundaries.

    Parameters
    ----------
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    *args
        Variable length argument list containing arrays to include in the
        domain decomposition.
    molecules : (N,) numpy.ndarray, optional
        Array of integer molecule affiliation for each of :code:`N` particles.
        Global (across all MPI ranks) or local (local indices on this MPI rank
        only) may be used, both, without affecting the result.
    bonds : (N,M) numpy.ndarray, optional
        Array of :code:`M` bonds originating from each of :code:`N` particles.
    verbose : int, optional
        Specify the logging event verbosity of this function.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    Returns
    -------
    Domain decomposed :code:`positions` and any array specified in
    :code:`*args`, in addition to :code:`bonds` and :code:`molecules` if these
    arrays were provided as input.
    """
    if molecules is not None:
        assert bonds is not None, "bonds must be provided with molecules"
        unique_molecules = np.sort(np.unique(molecules))
        molecules_com = np.empty_like(positions)
        for m in unique_molecules:
            ind = molecules == m
            r = positions[ind, :][0, :]
            molecules_com[ind, :] = r
        layout = pm.decompose(molecules_com, smoothing=0)
        args = (*args, bonds, molecules)
    else:
        layout = pm.decompose(positions, smoothing=0)
    if verbose > 1:
        Logger.rank0.log(
            logging.INFO,
            "DOMAIN_DECOMP: Total number of particles to be exchanged = %d",
            np.sum(layout.get_exchange_cost()),
        )
    return layout.exchange(positions, *args)



"""Handle file input/output in parllel HDF5 fashion
"""
import numpy as np
import h5py
import os
import logging
import getpass
from mpi4py import MPI
from .logger import Logger, get_version


class OutDataset:
    """HDF5 dataset handler for file output"""

    def __init__(
        self,
        dest_directory,
        config,
        double_out=False,
        disable_mpio=False,
        comm=MPI.COMM_WORLD,
    ):
        """Constructor

        Parameters
        ----------
        dest_directory : str
            Ouput directory for saving data.
        config : Config
            Configuration object.
        double_out : bool, optional
            If :code:`True`, the output HDF5 objects are written in eight byte
            floating point representation. Otherwise, four byte single
            precision is used.
        disable_mpio : bool, optional
            If :code:`True`, disables parallel MPI-enabled HDF5 file output.
            This is a compatibility option used if HDF5 is not compiled with
            MPI support. This makes everything much harder by splitting all
            output in one file per MPI rank, and having an MPI-enabled HDF5
            library is **highly recommended**.
        comm : mpi4py.Comm
            MPI communicator to use for rank commuication.
        """
        self.disable_mpio = disable_mpio
        self.config = config
        if double_out:
            self.float_dtype = "float64"
        else:
            self.float_dtype = "float32"

        if disable_mpio:
            self.file = h5py.File(
                os.path.join(
                    dest_directory, f"sim.hdf5-{comm.rank:6d}-of-{comm.size:6d}"
                ),
                "w",
            )
        else:
            self.file = h5py.File(
                os.path.join(dest_directory, "sim.H5"), "w", driver="mpio", comm=comm
            )

    def is_open(self, comm=MPI.COMM_WORLD):
        """Check if HDF5 output file is open

        Parameters
        ----------
        comm : mpi4py.Comm
            MPI communicator to use for rank communication.
        """
        comm.Barrier()
        return self.file.__bool__()

    def close_file(self, comm=MPI.COMM_WORLD):
        """Closes the HDF5 output file

        Parameters
        ----------
        comm : mpi4py.Comm
            MPI communicator to use for rank communication.
        """
        comm.Barrier()
        self.file.close()

    def flush(self):
        """Flushes output buffers, forcing file writes"""
        self.file.flush()


def setup_time_dependent_element(
    name, parent_group, n_frames, shape, dtype, units=None
):
    """Helper function for setting up time-dependent HDF5 group datasets

    All output groups must adhere to the H5MD standard, meaning a structure of


    |    ┗━ **group** particle group (e.g. :code:`all`) or :code:`observables` group
    |        ┗━ **group** time-dependent data
    |            ┣━ **dataset** :code:`step` :code:`shape=(n_frames,)`
    |            ┣━ **dataset** :code:`time` :code:`shape=(n_frames,)`
    |            ┗━ **dataset** :code:`value` :code:`shape=(n_frames, *)`

    is necessary.

    References
    ----------
    H5MD specification :
        https://www.nongnu.org/h5md/h5md.html
    """  # noqa: E501
    group = parent_group.create_group(name)
    step = group.create_dataset("step", (n_frames,), "int32")
    time = group.create_dataset("time", (n_frames,), "float32")
    value = group.create_dataset("value", (n_frames, *shape), dtype)

    if units is not None:
        value.attrs["unit"] = units
        time.attrs["unit"] = "ps"
    return group, step, time, value


def store_static(
    h5md,
    rank_range,
    names,
    types,
    indices,
    config,
    bonds_2_atom1,
    bonds_2_atom2,
    molecules=None,
    velocity_out=False,
    force_out=False,
    charges=False,
    dielectrics=False,
    plumed_out=False,
    comm=MPI.COMM_WORLD,
):
    """Outputs all static time-independent quantities to the HDF5 output file

    Parameters
    ----------
    h5md : OutDataset
        HDF5 dataset handler.
    rank_range : list[int]
        Start and end indices for global arrays for each MPI rank.
    names : (N,) numpy.ndarray
        Array of names for :code:`N` particles.
    types : (N,) numpy.ndarray
        Array of type indices for :code:`N` particles.
    indices : (N,) numpy.ndarray
        Array of indices for :code:`N` particles.
    config : Config
        Configuration object.
    bonds_2_atom1 : (B,) numpy.ndarray
        Array of indices of the first particle for :code:`B` total
        two-particle bonds.
    bonds_2_atom2 : (B,) numpy.ndarray
        Array of indices of the second particle for :code:`B` total
        two-particle bonds.
    molecules : (N,) numpy.ndarray, optional
        Array of integer molecule affiliation for each of :code:`N` particles.
        Global (across all MPI ranks) or local (local indices on this MPI rank
        only) may be used, both, without affecting the result.
    velocity_out : bool, optional
        If :code:`True`, velocities are written to output HDF5 file.
    force_out : bool, optional
        If :code:`True`, forces are written to output HDF5 file.
    charges : (N,) numpy.ndarray
        Array of particle charge values for :code:`N` particles.
    dielectrics : (N,) numpy.ndarray
        Array of particle relative dielectric values for :code:`N` particles.
    plumed_out : bool, optional
        If :code:`True`, PLUMED bias is written to output HDF5 file.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    See also
    --------
    prepare_bonds :
        Constructs two-, three-, and four-particle bonds from toplogy input
        file and bond configuration information.
    distribute_input :
        Distributes input arrays onto MPI ranks, attempting load balancing.
    """
    dtype = h5md.float_dtype

    h5md_group = h5md.file.create_group("/h5md")
    h5md.h5md_group = h5md_group
    h5md.observables = h5md.file.create_group("/observables")
    h5md.connectivity = h5md.file.create_group("/connectivity")
    h5md.parameters = h5md.file.create_group("/parameters")

    h5md_group.attrs["version"] = np.array([1, 1], dtype=int)
    author_group = h5md_group.create_group("author")
    author_group.attrs["name"] = np.string_(getpass.getuser())
    creator_group = h5md_group.create_group("creator")
    creator_group.attrs["name"] = np.string_("Hylleraas MD")

    # Get HyMD version. Also grab the user email from git config if we
    # can find it.
    creator_group.attrs["version"] = np.string_(get_version())
    try:
        import git

        try:
            reader = repo.config_reader()
            user_email = reader.get_value("user", "email")
            author_group.attrs["email"] = np.string_(user_email)
        except:
            pass
    except:
        pass

    h5md.particles_group = h5md.file.create_group("/particles")
    h5md.all_particles = h5md.particles_group.create_group("all")
    mass = h5md.all_particles.create_dataset("mass", (config.n_particles,), dtype)
    mass[...] = config.mass

    if charges is not False:
        charge = h5md.all_particles.create_dataset(
            "charge", (config.n_particles,), dtype="float32"
        )
        charge[indices] = charges
    if dielectrics is not False:
        dielectric = h5md.all_particles.create_dataset(
            "dielectric", (config.n_particles,), dtype="float32"
        )
        dielectric[indices] = dielectrics

    box = h5md.all_particles.create_group("box")
    box.attrs["dimension"] = 3
    box.attrs["boundary"] = np.array(
        [np.string_(s) for s in 3 * ["periodic"]], dtype="S8"
    )

    n_frames = config.n_steps // config.n_print
    if np.mod(config.n_steps - 1, config.n_print) != 0:
        n_frames += 1
    if np.mod(config.n_steps, config.n_print) == 1:
        n_frames += 1
    if n_frames == config.n_steps:
        n_frames += 1

    species = h5md.all_particles.create_dataset(
        "species",
        (config.n_particles,),
        dtype="i",
    )

    (
        _,
        h5md.positions_step,
        h5md.positions_time,
        h5md.positions,
    ) = setup_time_dependent_element(
        "position",
        h5md.all_particles,
        n_frames,
        (config.n_particles, 3),
        dtype,
        units="nm",
    )
    if velocity_out:
        (
            _,
            h5md.velocities_step,
            h5md.velocities_time,
            h5md.velocities,
        ) = setup_time_dependent_element(
            "velocity",
            h5md.all_particles,
            n_frames,
            (config.n_particles, 3),
            dtype,
            units="nm ps-1",
        )
    if force_out:
        (
            _,
            h5md.forces_step,
            h5md.forces_time,
            h5md.forces,
        ) = setup_time_dependent_element(
            "force",
            h5md.all_particles,
            n_frames,
            (config.n_particles, 3),
            dtype,
            units="kJ mol-1 nm-1",
        )
    (
        _,
        h5md.total_energy_step,
        h5md.total_energy_time,
        h5md.total_energy,
    ) = setup_time_dependent_element(
        "total_energy",
        h5md.observables,
        n_frames,
        (1,),
        dtype,
        units="kJ mol-1",  # noqa: E501
    )
    (
        _,
        h5md.kinetc_energy_step,
        h5md.kinetc_energy_time,
        h5md.kinetc_energy,
    ) = setup_time_dependent_element(
        "kinetic_energy",
        h5md.observables,
        n_frames,
        (1,),
        dtype,
        units="kJ mol-1",  # noqa: E501
    )
    (
        _,
        h5md.potential_energy_step,
        h5md.potential_energy_time,
        h5md.potential_energy,
    ) = setup_time_dependent_element(
        "potential_energy",
        h5md.observables,
        n_frames,
        (1,),
        dtype,
        units="kJ mol-1",  # noqa: E501
    )
    (
        _,
        h5md.bond_energy_step,
        h5md.bond_energy_time,
        h5md.bond_energy,
    ) = setup_time_dependent_element(
        "bond_energy",
        h5md.observables,
        n_frames,
        (1,),
        dtype,
        units="kJ mol-1",  # noqa: E501
    )
    (
        _,
        h5md.angle_energy_step,
        h5md.angle_energy_time,
        h5md.angle_energy,
    ) = setup_time_dependent_element(
        "angle_energy",
        h5md.observables,
        n_frames,
        (1,),
        dtype,
        units="kJ mol-1",  # noqa: E501
    )
    (
        _,
        h5md.dihedral_energy_step,
        h5md.dihedral_energy_time,
        h5md.dihedral_energy,
    ) = setup_time_dependent_element(
        "dihedral_energy",
        h5md.observables,
        n_frames,
        (1,),
        dtype,
        units="kJ mol-1",  # noqa: E501
    )
    (
        _,
        h5md.field_energy_step,
        h5md.field_energy_time,
        h5md.field_energy,
    ) = setup_time_dependent_element(
        "field_energy",
        h5md.observables,
        n_frames,
        (1,),
        dtype,
        units="kJ mol-1",  # noqa: E501
    )
    if charges is not False:
        (
            _,
            h5md.field_q_energy_step,
            h5md.field_q_energy_time,
            h5md.field_q_energy,
        ) = setup_time_dependent_element(
            "field_q_energy",
            h5md.observables,
            n_frames,
            (1,),
            dtype,
            units="kJ mol-1",  # noqa: E501
        )  # <-------- xinmeng
    if plumed_out is not False:
        (
            _,
            h5md.plumed_bias_step,
            h5md.plumed_bias_time,
            h5md.plumed_bias,
        ) = setup_time_dependent_element(
            "plumed_bias", h5md.observables, n_frames, (1,), dtype, units="kJ mol-1"  # noqa: E501
        )
    (
        _,
        h5md.total_momentum_step,
        h5md.total_momentum_time,
        h5md.total_momentum,
    ) = setup_time_dependent_element(  # noqa: E501
        "total_momentum",
        h5md.observables,
        n_frames,
        (3,),
        dtype,
        units="nm g ps-1 mol-1",
    )
    (
        _,
        h5md.angular_momentum_step,
        h5md.angular_momentum_time,
        h5md.angular_momentum,
    ) = setup_time_dependent_element(  # noqa: E501
        "angular_momentum",
        h5md.observables,
        n_frames,
        (3,),
        dtype,
        units="nm+2 g ps-1 mol-1",
    )
    (
        _,
        h5md.torque_step,
        h5md.torque_time,
        h5md.torque,
    ) = setup_time_dependent_element(  # noqa: E501
        "torque",
        h5md.observables,
        n_frames,
        (3,),
        dtype,
        units="kJ nm+2 mol-1",
    )
    (
        _,
        h5md.temperature_step,
        h5md.temperature_time,
        h5md.temperature,
    ) = setup_time_dependent_element(
        "temperature", h5md.observables, n_frames, (3,), dtype, units="K"
    )
    (
        _,
        h5md.thermostat_work_step,
        h5md.thermostat_work_time,
        h5md.thermostat_work,
    ) = setup_time_dependent_element(
        "thermostat_work",
        h5md.observables,
        n_frames,
        (1,),
        "float32",
        units="kJ mol-1",  # noqa: E501
    )
    (
        _,
        h5md.pressure_step,
        h5md.pressure_time,
        h5md.pressure,
    ) = setup_time_dependent_element(
        "pressure", h5md.observables, n_frames, (18,), "float32", units="Bar"
    )
    (_, h5md.box_step, h5md.box_time, h5md.box_value,) = setup_time_dependent_element(
        "edges", box, n_frames, (3, 3), "float32", units="nm"
    )

    ind_sort = np.argsort(indices)
    for i in ind_sort:
        species[indices[i]] = config.name_to_type_map[names[i].decode("utf-8")]

    h5md.parameters.attrs["config.toml"] = np.string_(str(config))
    vmd_group = h5md.parameters.create_group("vmd_structure")
    index_of_species = vmd_group.create_dataset(
        "indexOfSpecies", (config.n_types,), "i"
    )
    index_of_species[:] = np.array(list(range(config.n_types)))

    # VMD-h5mdplugin maximum name/type name length is 16 characters (for
    # whatever reason [VMD internals?]).
    name_dataset = vmd_group.create_dataset("name", (config.n_types,), "S16")
    type_dataset = vmd_group.create_dataset("type", (config.n_types,), "S16")
    if molecules is not None:
        resid_dataset = vmd_group.create_dataset(
            "resid",
            (config.n_particles,),
            "i",
        )

    # Change this
    for i, n in config.type_to_name_map.items():
        name_dataset[i] = np.string_(n[:16])
        if n == "W":
            type_dataset[i] = np.string_("solvent")
        else:
            type_dataset[i] = np.string_("membrane")

    total_bonds = comm.allreduce(len(bonds_2_atom1), MPI.SUM)
    n_bonds_local = len(bonds_2_atom1)

    receive_buffer = comm.gather(n_bonds_local, root=0)
    n_bonds_global = None
    if comm.Get_rank() == 0:
        n_bonds_global = receive_buffer
    n_bonds_global = np.array(comm.bcast(n_bonds_global, root=0))
    rank_bond_start = np.sum(n_bonds_global[: comm.Get_rank()])

    bonds_from = vmd_group.create_dataset("bond_from", (total_bonds,), "i")
    bonds_to = vmd_group.create_dataset("bond_to", (total_bonds,), "i")
    for i in range(n_bonds_local):
        a = bonds_2_atom1[i]
        b = bonds_2_atom2[i]
        bonds_from[rank_bond_start + i] = indices[a] + 1
        bonds_to[rank_bond_start + i] = indices[b] + 1

    if molecules is not None:
        resid_dataset[indices[ind_sort]] = molecules


# store data old vs
"""
h5md, step, frame, indices, positions, velocities, forces, box_size,
temperature, kinetic_energy, bond2_energy, bond3_energy, bond4_energy,
field_energy, field_q_energy, time_step, config, velocity_out=False,
force_out=False, charge_out=False, dump_per_particle=False,
"""


def store_data(
    h5md,
    step,
    frame,
    indices,
    positions,
    velocities,
    forces,
    box_size,
    temperature,
    pressure,
    kinetic_energy,
    bond2_energy,
    bond3_energy,
    bond4_energy,
    field_energy,
    field_q_energy,
    plumed_bias,
    time_step,
    config,
    velocity_out=False,
    force_out=False,
    charge_out=False,
    plumed_out=False,
    dump_per_particle=False,
    comm=MPI.COMM_WORLD,
):
    """Writes time-step data to HDF5 output file

    Handles all quantities which change during simulation, as opposed to
    static quanitities (see :code:`store_static`).

    Parameters
    ----------
    h5md : OutDataset
        HDF5 dataset handler.
    step : int
        Step number.
    frame : int
        Output frame number (:code:`step // n_print`).
    indices : (N,) numpy.ndarray
        Array of indices for :code:`N` particles.
    positions : (N,) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
    velocities : (N,) numpy.ndarray
        Array of velocities for :code:`N` particles in :code:`D` dimensions.
    forces : (N,) numpy.ndarray
        Array of forces for :code:`N` particles in :code:`D` dimensions.
    box_size : (D,) numpy.ndarray
        Array containing the simulation box size.
    temperature : float
        Calculated instantaneous temperature.
    kinetic_energy : float
        Calculated instantaneous kinetic energy.
    bond2_energy : float
        Calculated instantaneous harmonic two-particle bond energy.
    bond3_energy : float
        Calculated instantaneous harmonic angular three-particle bond energy.
    bond4_energy : float
        Calculated instantaneous dihedral four-particle torsion energy.
    field_energy : float
        Calculated instantaneous particle-field energy.
    field_q_energy : float
        Calculated instantaneous electrostatic energy.
    plumed_bias : float
        PLUMED instantaneous bias energy.
    time_step : float
        Value of the time step.
    config : Config
        Configuration object.
    velocity_out : bool, optional
        If :code:`True`, velocities are written to output HDF5 file.
    force_out : bool, optional
        If :code:`True`, forces are written to output HDF5 file.
    charge_out : bool, optional
        If :code:`True`, electrostatic energies are written to the output
        HDF5 file.
    plumed_out : bool, optional
        If :code:`True`, PLUMED bias is written to the output HDF5 file.
    dump_per_particle : bool, optional
        If :code:`True`, all quantities are written **per particle**.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    See also
    --------
    store_static :
        Outputs all static time-independent quantities to the HDF5 output file
    """
    for dset in (
        h5md.positions_step,
        h5md.total_energy_step,
        h5md.potential_energy_step,
        h5md.kinetc_energy_step,
        h5md.bond_energy_step,
        h5md.angle_energy_step,
        h5md.dihedral_energy_step,
        h5md.field_energy_step,
        h5md.total_momentum_step,
        h5md.angular_momentum_step,
        h5md.torque_step,
        h5md.temperature_step,
        h5md.pressure_step,
        h5md.box_step,
        h5md.thermostat_work_step,
    ):
        dset[frame] = step

    for dset in (
        h5md.positions_time,
        h5md.total_energy_time,
        h5md.potential_energy_time,
        h5md.kinetc_energy_time,
        h5md.bond_energy_time,
        h5md.angle_energy_time,
        h5md.dihedral_energy_time,
        h5md.field_energy_time,
        h5md.total_momentum_time,
        h5md.angular_momentum_time,
        h5md.torque_time,
        h5md.temperature_time,
        h5md.pressure_time,
        h5md.box_time,
        h5md.thermostat_work_time,
    ):
        dset[frame] = step * time_step

    if velocity_out:
        h5md.velocities_step[frame] = step
        h5md.velocities_time[frame] = step * time_step
    if force_out:
        h5md.forces_step[frame] = step
        h5md.forces_time[frame] = step * time_step
    if charge_out:
        h5md.field_q_energy_step[frame] = step
        h5md.field_q_energy_time[frame] = step * time_step
    if plumed_out:
        h5md.plumed_bias_step[frame] = step
        h5md.plumed_bias_time[frame] = step * time_step

    ind_sort = np.argsort(indices)
    h5md.positions[frame, indices[ind_sort]] = positions[ind_sort]

    if velocity_out:
        h5md.velocities[frame, indices[ind_sort]] = velocities[ind_sort]
    if force_out:
        h5md.forces[frame, indices[ind_sort]] = forces[ind_sort]
    if charge_out:
        h5md.field_q_energy[frame] = field_q_energy
    if plumed_out:
        h5md.plumed_bias[frame] = plumed_bias

    potential_energy = (
        bond2_energy + bond3_energy + bond4_energy + field_energy + field_q_energy
    )

    total_momentum = config.mass * comm.allreduce(np.sum(velocities, axis=0), MPI.SUM)
    angular_momentum = config.mass * comm.allreduce(
        np.sum(np.cross(positions, velocities), axis=0), MPI.SUM
    )
    torque = config.mass * comm.allreduce(
        np.sum(np.cross(positions, forces), axis=0), MPI.SUM
    )
    h5md.total_energy[frame] = kinetic_energy + potential_energy
    h5md.potential_energy[frame] = potential_energy
    h5md.kinetc_energy[frame] = kinetic_energy
    h5md.bond_energy[frame] = bond2_energy
    h5md.angle_energy[frame] = bond3_energy
    h5md.dihedral_energy[frame] = bond4_energy
    h5md.field_energy[frame] = field_energy
    h5md.total_momentum[frame, :] = total_momentum
    h5md.angular_momentum[frame, :] = angular_momentum
    h5md.torque[frame, :] = torque
    h5md.temperature[frame] = temperature
    h5md.pressure[frame] = pressure
    for d in range(3):
        h5md.box_value[frame, d, d] = box_size[d]
    h5md.thermostat_work[frame] = config.thermostat_work

    fmt_ = [
        "step",
        "time",
        "temp",
        "tot E",
        "kin E",
        "pot E",
        "field E",
        "elec E",
        "bond E",
        "ang E",
        "dih E",
        "bias E",
        "Px",
        "Py",
        "Pz",
        "ΔH" if config.target_temperature else "ΔE",
    ]
    fmt_ = np.array(fmt_)

    # create mask to show only energies != 0
    en_array = np.array([
        field_energy,
        field_q_energy,
        bond2_energy,
        bond3_energy,
        bond4_energy,
        plumed_bias,
    ])
    mask = np.full_like(fmt_, True, dtype=bool)
    mask[range(6,12)] = en_array != 0.

    header_ = fmt_[mask].shape[0] * "{:>13}"
    if config.initial_energy is None:
        fmt_[-1] = ""

    divide_by = 1.0
    if dump_per_particle:
        for i in range(3, 9):
            fmt_[i] = fmt_[i][:-2] + "E/N"
        fmt_[-1] += "/N"
        divide_by = config.n_particles
    total_energy = kinetic_energy + potential_energy
    if config.initial_energy is not None:
        if config.target_temperature:
            H_tilde = total_energy - config.initial_energy - config.thermostat_work
        else:
            H_tilde = total_energy - config.initial_energy
    else:
        H_tilde = 0.0

    header = header_.format(*fmt_[mask])
    data_fmt = f'{"{:13}"}{(fmt_[mask].shape[0]-1) * "{:13.5g}" }'
    all_data = [
        step,
        time_step * step,
        temperature,
        total_energy / divide_by,
        kinetic_energy / divide_by,
        potential_energy / divide_by,
        field_energy / divide_by,
        field_q_energy / divide_by,
        bond2_energy / divide_by,
        bond3_energy / divide_by,
        bond4_energy / divide_by,
        plumed_bias / divide_by,
        total_momentum[0] / divide_by,
        total_momentum[1] / divide_by,
        total_momentum[2] / divide_by,
        H_tilde / divide_by,
    ]
    data = data_fmt.format(*[val for i, val in enumerate(all_data) if mask[i]])
    Logger.rank0.log(logging.INFO, ("\n" + header + "\n" + data))


def distribute_input(
    in_file, rank, size, n_particles, max_molecule_size=201, comm=MPI.COMM_WORLD
):
    """Assign global arrays onto MPI ranks, attempting load balancing

    Distributes approximately equal numbers of particles (workload) onto each
    independent MPI rank, while respecting the requirement that any molecule
    must be fully contained on a single MPI rank only (no splitting molecules
    across multiple CPUs).

    Parameters
    ----------
    in_file : h5py.File
        HDF5 input file object.
    rank : int
        Local rank number for this MPI rank.
    size : int
        Global size of the MPI communicator (number of total CPUs).
    n_particles : int
        Total number of particles.
    max_molecule_size : int, optional
        Maximum size of any molecule present in the system. Used to initially
        guess where the MPI rank boundaries (start/end indices) in the global
        arrays should be placed. If molecules of size
        :code:`>max_molecule_size` exist in the simulation system, HyMD
        **might** work as expected. Or it might fail spectacularly.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    Returns
    -------
    rank_range :
        Starting and ending indices in the global arrays for each MPI rank.
    """
    if n_particles is None:
        n_particles = len(in_file["indices"])
    np_per_MPI = n_particles // size

    molecules_flag = False
    if "molecules" in in_file:
        molecules_flag = True

    if not molecules_flag:
        if rank == size - 1:
            np_cum_mpi = [rank * np_per_MPI, n_particles]
        else:
            np_cum_mpi = [rank * np_per_MPI, (rank + 1) * np_per_MPI]
        p_mpi_range = list(range(np_cum_mpi[0], np_cum_mpi[1]))
        return p_mpi_range, molecules_flag

    # To avoid splitting molecules across multiple different ranks, we need
    # to read in some extra indices before/after the expected break points
    # and iterate until we find a molecule break.
    #
    # Implicitly assuming no molecule is bigger than
    # min(max_molecule_size, n_particles // n_MPI_ranks) atoms.
    max_molecule_size += 2
    grab_extra = max_molecule_size if np_per_MPI > max_molecule_size else np_per_MPI
    if rank == 0:
        mpi_range_start = 0
        if size == 1:
            mpi_range_end = n_particles
        else:
            mpi_range_end = (rank + 1) * np_per_MPI + grab_extra
    elif rank == size - 1:
        mpi_range_start = rank * np_per_MPI - 1
        mpi_range_end = n_particles
    else:
        mpi_range_start = rank * np_per_MPI - 1
        mpi_range_end = (rank + 1) * np_per_MPI + grab_extra

    molecules = in_file["molecules"][mpi_range_start:mpi_range_end]
    indices = in_file["indices"][mpi_range_start:mpi_range_end]
    molecule_end_indices = np.nonzero(np.diff(molecules))[0]

    p_mpi_range = [None, None]
    if rank == 0:
        p_mpi_range[0] = 0
        if size == 1:
            p_mpi_range[1] = n_particles
        else:
            p_mpi_range[1] = indices[
                molecule_end_indices[molecule_end_indices >= np_per_MPI][0] + 1
            ]
    elif rank == size - 1:
        p_mpi_range[0] = indices[molecule_end_indices[molecule_end_indices > 0][0]] + 1
        p_mpi_range[1] = n_particles
    else:
        p_mpi_range[0] = indices[molecule_end_indices[molecule_end_indices > 0][0]] + 1
        p_mpi_range[1] = (
            indices[molecule_end_indices[molecule_end_indices > np_per_MPI][0]]
            + 1  # noqa: E501
        )
    return list(range(p_mpi_range[0], p_mpi_range[1])), molecules_flag
